package org.cakelab.codegen;

import java.io.IOException;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;


public abstract class ClassGen implements Runnable {

	protected SourceCodeUnit sourceCodeUnit;
	protected String pkgName;
	protected String className;
	protected String modifiers = "public";

	protected String superClass;
	protected HashMap<String, String> implementedInterfaces = new HashMap<String, String>();
	protected StaticInitBlock staticBlock = null;
	protected HashMap<String, ClassGen> memberClasses = new HashMap<String, ClassGen>();
	protected HashMap<String, MethodGen> methods = new HashMap<String, MethodGen>();
	protected HashMap<String, AttributeGen> attributes = new HashMap<String, AttributeGen>();
	private JavaDoc javadoc;

	public ClassGen(SourceCodeUnit sourceCodeUnit, String modifiers, String pkg, String className, String superClassExpression, Class<?> ... interfaces) {
		this.sourceCodeUnit = sourceCodeUnit;
		this.modifiers = modifiers;
		this.pkgName = pkg;
		this.className = className;
		this.superClass = superClassExpression;
		for (Class<?> interfaze : interfaces) addInterface(interfaze);
		
	}

	public void setJavaDoc(JavaDoc javadoc) {
		this.javadoc = javadoc;
	}

	public void setStaticBlock(StaticInitBlock staticInitBlock) {
		this.staticBlock = staticInitBlock;
	}
	
	public void addInterface(Class<?> interfaze) {
		addInterface(interfaze.getCanonicalName(), interfaze.getCanonicalName());
	}

	public void addInterface(String canonicalName, String interfaceExpression) {
		if (!implementedInterfaces.containsKey(canonicalName)) {
			sourceCodeUnit.addImport(canonicalName);
			implementedInterfaces.put(canonicalName, interfaceExpression);
		}
	}

	/**
	 * This method is supposed to be implemented by derived classes to 
	 * populate the class body with members.
	 */
	protected abstract void populateBody();

	protected void addMethod(MethodGen methodGen) {
		methods.put(methodGen.getSignature(), methodGen);
	}

	protected void addField(AttributeGen attrGen) {
		attributes.put(attrGen.getName(), attrGen);
	}

	public void write(IndentStreamControl pout) throws IOException {
		
		if (this.javadoc != null) javadoc.write(pout);
		
		pout.print(modifiers + " class " + className);

		if (superClass != null) {
			pout.print(" extends " + superClass);
		}
		
		int num = implementedInterfaces.size();
		if (num > 0) {
			pout.print(" implements ");
			Iterator<String> i = implementedInterfaces.values().iterator();
			while (i.hasNext()) {
				pout.print(i.next());
				if (i.hasNext()) pout.print(", ");
			}
		}
		
		pout.println(" {");
		pout.indent(+1);
		
		
		if (this.staticBlock != null) {
			staticBlock.write(pout);
		}
		
		for (ClassGen memberClass : memberClasses.values()) {
			memberClass.write(pout);
			pout.println();
		}
		
		AttributeGen[] sortedAttributes = attributes.values().toArray(new AttributeGen[0]);
		Arrays.sort(sortedAttributes, MemberGen.COMPERATOR);
		for (AttributeGen attr : sortedAttributes) {
			attr.write(pout);
		}
		pout.println();

		MethodGen[] sortedMethods = methods.values().toArray(new MethodGen[0]); 
		Arrays.sort(sortedMethods, MethodGen.COMPERATOR);
		
		for (MethodGen method : sortedMethods) {
			method.write(pout);
			pout.println();
		}
		
		
		pout.indent(-1);
		pout.println("}");
	}

	
	public String getPackageName() {
		return pkgName;
	}

	public String getClassName() {
		return className;
	}
	
	public String getCanonicalName() {
		return this.pkgName + "." + className;
	}

	public void addImport(String name) {
		sourceCodeUnit.addImport(name);
	}


}
