package org.cakelab.codegen;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;

public class IndentStreamControl extends PrintStream {

	public static class IndentingStream extends OutputStream {

		/** output stream, which receives the indented output */
		private OutputStream out;
		/** control which controls the level of indent */
		private IndentStreamControl control;
		/** string representing a single indent level */
		private String indentString;
		/** buffer which holds a line (or fragment of it) without indent */
		private ByteArrayOutputStream line;
		/** indicates, whether there was a newline recently */
		private boolean newline;

		/**
		 * @param out stream, which receives written (and indented) bytes
		 * @param indentString String used for indenting. The resulting indent will be number of indent levels times this string.
		 */
		public IndentingStream(OutputStream out, String indentString) {
			this.out = out;
			this.indentString = indentString;
			this.newline = true;
			this.line = new ByteArrayOutputStream();
		}

		@Override
		public void write(int b) throws IOException {
			line.write(b);
			if (b == '\n') {
				endln();
			}
		}
		
		/** called on each detected new line symbol '\n' */
		private void endln() throws IOException {
			flush();
			newline = true;
		}

		@Override
		public void flush() throws IOException {
			// Only add indent, if the last character written was a new line
			// and the buffer is not empty!
			if (newline && line.size() > 0) {
				for (int i = control.getCurrentIndent(); i > 0; i--) {
					out.write(indentString.getBytes());
				}
				// we've processed the last seen new line character.
				newline = false;
			}
			out.write(line.toByteArray());
			line.reset();
		}

		public void setControl(IndentStreamControl control) {
			this.control = control;
		}

	}

	private static IndentingStream indenter;
	private int indent;
	private boolean ignoreClose = false;

	public IndentStreamControl(OutputStream out) {
		this(out, "    ");
	}
	
	public IndentStreamControl(OutputStream out, String indentString) {
		super(indenter = new IndentingStream(out, indentString), false);
		indenter.setControl(this);
		if (out == System.out || out == System.err) {
			ignoreClose = true;
		}
	}

	public int getCurrentIndent() {
		return indent;
	}

	/**
	 * Sets indent to given v.
	 * @see #indent(int)
	 * @param v
	 */
	public void setIndent(int v) {
		indent = v;
	}
	
	/**
	 * Increments indent by v (which can be negative)
	 * 
	 * Change will affect current line and following lines.
	 * @param v
	 * @throws IOException 
	 */
	public void indent(int v) {
		indent += v;
	}
	
	/**
	 * Indent stream will ignore calls to this method if the stream is either System.out or System.err .
	 */
	public void close() {
		if (ignoreClose) flush();
		else super.close();
	}
}
