package org.cakelab.codegen.compiler;

import java.io.Closeable;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.Arrays;
import java.util.Locale;

import javax.tools.Diagnostic;
import javax.tools.DiagnosticListener;
import javax.tools.JavaCompiler;
import javax.tools.JavaFileObject;
import javax.tools.StandardJavaFileManager;
import javax.tools.ToolProvider;

public class Compiler implements Closeable, DiagnosticListener<JavaFileObject> {
	private JavaCompiler compiler;
	private StandardJavaFileManager fileManager;
	private File tmpDir;
	private boolean useFiles;

	public Compiler (File classFileCache, boolean useFiles) throws IOException {
        this.useFiles = useFiles;
		compiler = ToolProvider.getSystemJavaCompiler();
        fileManager = compiler.getStandardFileManager(null, null, null);
		tmpDir = classFileCache;
		if (!tmpDir.exists()) throw new IOException("need working directory to cache class files at: " + tmpDir.getCanonicalPath());
		if (!tmpDir.canWrite()) throw new IOException("need write access to class file cache at: " + tmpDir.getCanonicalPath());
	}

	@Override
	public void close() throws IOException {
        fileManager.close();
	}
	
	
	public void compile(String classFqN, String sourceCode) throws IOException {
		Iterable<? extends JavaFileObject> compilationUnits;
		if (useFiles) {
			String name = classFqN.replace('.','/') + ".java";
			File f = new File(tmpDir, name);
			if (f.exists()) throw new IOException(f.toString() + " exists");
			f.getParentFile().mkdirs();
			
			
			FileOutputStream out = new FileOutputStream(f);
			PrintStream pout = new PrintStream(out);
			pout.print(sourceCode);
			pout.flush();
			pout.close();
			
	        compilationUnits = fileManager.getJavaFileObjectsFromFiles(Arrays.asList(f));

		} else {
			JavaSourceFromString compUnit = new JavaSourceFromString(classFqN, sourceCode);
			compilationUnits = Arrays.asList(compUnit);
		}
		
		compile(compilationUnits);

	}
	
	public void compile(Iterable<? extends JavaFileObject> compilationUnits) {
		compiler.getTask(null, fileManager, this, null, null, compilationUnits).call();
	}

	public void canCompile(String name) throws IOException {
		if (useFiles) {
			name = name.replace('.','/') + ".class";
			File f = new File(tmpDir, name);
			if (f.exists()) throw new IOException(f.toString() + " exists");
		}
	}

	public StandardJavaFileManager getFileManager() {
		return fileManager;
	}

	@Override
	public void report(Diagnostic<? extends JavaFileObject> diagnostic) {
		throw new Error(diagnostic.getKind() + ":" + diagnostic.getSource().getName() + ":" + diagnostic.getLineNumber() + ":" + diagnostic.getColumnNumber() + ": " + diagnostic.getMessage(Locale.getDefault()));
	}


}
