package org.cakelab.nio.generator;

import java.io.Closeable;
import java.io.File;
import java.io.IOException;
import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Queue;

import javax.tools.JavaFileObject;
import javax.tools.StandardJavaFileManager;

import org.cakelab.codegen.SourceCodeUnit;
import org.cakelab.codegen.compiler.Compiler;
import org.cakelab.nio.common.Global;
import org.cakelab.nio.generator.codegen.ViewClassGen;
import org.cakelab.nio.generator.codegen.struct.StructViewClassGen;
import org.cakelab.nio.view.GenericDataView;

public class Generator implements Closeable {
	
	static {
		Global.init();
	}


	
	public static final String BASE_PACKAGE = GenericDataView.class.getPackage().getName() + ".generated";
	
	public static String getStructViewClassFqN(Class<?> structType) {
		if (structType.isMemberClass()) {
			return getStructViewClassFqN(structType.getEnclosingClass()) + "." + getStructViewSimpleClassName(structType);
		} else {
			return getStructViewPackageName(structType) + '.' + getStructViewSimpleClassName(structType);
		}
	}

	public static String getStructViewSimpleClassName(Class<?> structType) {
		return structType.getSimpleName() + "View";
	}
	
	public static String getStructViewPackageName(Class<?> structType) {
		return BASE_PACKAGE + "." + structType.getPackage().getName();
	}

	
	/* ********************************************************************************************************** *
	 * ********************************************************************************************************** */
	private Compiler compiler;
	private File outDir;
	private boolean writeSourceFiles;
	private StandardJavaFileManager fileManager;
	private Queue<Runnable> tasks = new ArrayDeque<Runnable>();
	private HashSet<String> classes = new HashSet<String>();
	private ArrayList<JavaFileObject> cunits = new ArrayList<JavaFileObject>();
	private boolean alwaysOverwrite;
	private boolean compile;
	/* ********************************************************************************************************** *
	 * ********************************************************************************************************** */
	
	public Generator(File classFileCache, boolean writeSourceCodeFiles, boolean alwaysOverwrite, boolean compile) throws IOException {
		this.outDir = classFileCache;
		this.writeSourceFiles = writeSourceCodeFiles;
		this.alwaysOverwrite = alwaysOverwrite;
		this.compile = compile;
		this.compiler = new Compiler(classFileCache, writeSourceCodeFiles);
		this.fileManager = compiler.getFileManager();
	}

	public void close() throws IOException {
        compiler.close();
	}
	
	/**
	 * Generates and compiles a view for the given struct class and 
	 * structs referenced by the struct.
	 * 
	 * @param dataType
	 * @throws IOException
	 */
	public void generate(Class<?> dataType) throws IOException {
		
		scheduleDataViewGeneration(dataType);
		
		processGenerationTasks(tasks);

		if (compile) compileAll();
	}

	private void compileAll() {
		// compiles all compilation units that have been registered by 
		// class generation tasks and clears the list.
		compiler.compile(cunits);
		cunits.clear();
	}

	private void processGenerationTasks(Queue<Runnable> tasks) {
		// Processes all class generation tasks.
		// Class generators may schedule new tasks while processing.
		
		while (!tasks.isEmpty()) {
			Runnable t = tasks.remove();
			t.run();
		}
	}

	public void scheduleDataViewGeneration(Class<?> dataType) {
		// This method schedules a class generation task, which is processed later.
		// However, the class is also registered through this method, and subsequent 
		// requests to schedule class generation for the same class will be ignored.
		String pkgName;
		String className;
		String classFqN;
		
		pkgName = Generator.getStructViewPackageName(dataType);
		className = Generator.getStructViewSimpleClassName(dataType);
		classFqN = Generator.getStructViewClassFqN(dataType);
		
		if (!classes.contains(classFqN)) {
			Global.info("scheduling generation of data view for " + dataType.getCanonicalName());
			classes.add(classFqN);
			SourceCodeUnit code = new SourceCodeUnit(this, outDir, writeSourceFiles);
			ViewClassGen classGen;
			classGen = new StructViewClassGen(this, code, pkgName, className, dataType);
			code.setClassGen(classGen);
			tasks.add(code);
		}

	}
	
	public boolean checkTimeStamp(Class<?> structType, long startTime) {
		// TODO: dependency management and time stamp check
		if (alwaysOverwrite) return true;
		
		String classFqN = getStructViewClassFqN(structType);
		String name = classFqN.replace('.','/') + ".class";
		File classFile = new File(outDir, name);
		
		if (!classFile.exists()) {
			return true;
		} else {
			return classFile.lastModified() < startTime;
		}
	}

	public StandardJavaFileManager getFileManager() {
		return fileManager;
	}

	public void addCompilationUnit(JavaFileObject cunit) {
		cunits.add(cunit);
	}
	
}
