package org.cakelab.nio.generator;

import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.util.Arrays;

import org.cakelab.nio.Init;
import org.cakelab.nio.common.Global;

public class Main {
	static class Option<T> {
		boolean optional = true;
		Class<T> type;
		String[] names;
		String description;
		T value;
		@SuppressWarnings("unchecked")
		Option(String description, T defaultValue, String ... names) {
			this.optional = true;
			this.type = (Class<T>) defaultValue.getClass();
			this.names = names;
			this.description = description;
			this.value = defaultValue;
		}
		
		Option(String description, Class<T> type, String ... names) {
			this.optional = false;
			this.type = type;
			this.names = names;
			this.description = description;
			this.value = null;
		}
		
		boolean equals(String option) {
			for (String n : names) {
				if (option.equals(n)) {
					return true;
				}
			}
			return false;
		}
		
		@SuppressWarnings("unchecked")
		void setValue(String value) {
			if (type.equals(File.class)) {
				this.value = (T) new File(value);
			} else if (Number.class.isAssignableFrom(type)) {
				this.value = (T) Double.valueOf(value);
			} else if (Boolean.class == type) {
				this.value = (T) Boolean.valueOf(value);
			} else {
				throw new Error("you need to implement a conversion method");
			}
		}

		public void help(PrintStream print) {
			print.println(Arrays.toString(names));
			print.println("\t" + description);
			print.println("\tType: " + type.getSimpleName());
			if (optional && value != null) print.println("\tDefault: " + value.toString());
		}
	}
	
	static Option<File> out = new Option<File>("Output folder for generated java source files", 
									File.class, "o", "out");
	static Option<Boolean> source = new Option<Boolean>("Write source files for generated classes", Boolean.TRUE, "s", "source");
	static Option<Boolean> compile = new Option<Boolean>("Compile generated classes", Boolean.TRUE, "c", "compile");
	static Option<Boolean> replace = new Option<Boolean>("Always replace existing class files", Boolean.TRUE, "r", "replace");
	static Option<Boolean> version = new Option<Boolean>("Print version and exit", Boolean.FALSE, "v", "version");
	
	
	static Option<?>[] options = new Option<?>[]{ out, source, compile, replace, version };
	
	static Class<?>[] structs;
	
	public static void main(String[] args) {
		int i = 0;
		for (; i < args.length; i++) {
			String arg = args[i];
			if (!arg.startsWith("-")) {
				break;
			} 
			String option = arg.replaceFirst("-+", "");
			if (out.equals(option)) {
				out.setValue(args[++i]);
			} else if (version.equals(option)) {
				version.value = true;
			} else if (source.equals(option)) {
				source.setValue(args[++i]);
			} else if (compile.equals(option)) {
				compile.setValue(args[++i]);
			} else if (replace.equals(option)) {
				replace.setValue(args[++i]);
			} else {
				System.err.println("ERROR: unknown option " + option);
			}
		}

		if (!valid()) {
			System.err.flush();
			System.out.flush();
			printHelp(System.out);
			System.exit(-1);
		}
		
		int result;
		if (version.value) {
			if (checkVersion()) {
				System.out.println(Global.VERSION_STR);
				result = 0;
			} else {
				result = 128;
			}
		} else {
			String[] structs = Arrays.copyOfRange(args, i, args.length);
			result = execute(out.value, source.value, replace.value, compile.value, structs);
		}
		System.exit(result);
	}

	private static boolean checkVersion() {
		String versionTestAttrib = Global.VERSION_TEST_FIELD;
		try {
			Global.class.getField(versionTestAttrib);
		} catch (NoSuchFieldException | SecurityException e) {
			System.err.println("Inconsistent set of version specifiers in " + Global.class.getCanonicalName() + ".");
			System.err.println("Implementation requires version test field \n\t" + versionTestAttrib);
			System.err.println("to indicate compatibility with its own version \n\t" + Global.VERSION_STR);
			System.err.println(".");
			return false;
		}
		return true;
	}

	public static int execute(File outFolder, boolean source, boolean override, boolean compile, String[] structs) {
		Init.debug(Init.isAssertionsEnabled());
		
		if (!Init.isAssertionsEnabled()) {
			Global.info("assertions are disabled!");
		}
		
		Generator generator = null;
		try {
			generator = new Generator(outFolder, source, override, compile);
		} catch (IOException e1) {
			e1.printStackTrace();
			return -1;
		}
		for (String struct : structs) {
			try {
				generator.generate(loadClass(struct));
			} catch (ClassNotFoundException | IOException e) {
				Global.error("can't find class with name '" + struct + "' in the given classpath.");
				try {if (generator != null) generator.close();} catch(IOException ex){};
				return -1;
			}
		}
		try {
			generator.close();
		} catch (IOException e) {
			// who cares
		}
		return 0;
	}

	private static Class<?> loadClass(String arg) throws ClassNotFoundException {
		return Main.class.getClassLoader().loadClass(arg);
	}

	private static void printHelp(PrintStream print) {
		print.println("Java Native I/O Generator (" + Global.VERSION_STR + ")");
		for (Option<?> o : options) {
			o.help(print);
		}
	}

	private static boolean valid() {
		return version.value || (out.value != null && structs != null && structs.length > 0);
	}

}
