package org.cakelab.nio.generator.codegen;

import java.lang.reflect.Constructor;
import java.lang.reflect.Parameter;

import org.cakelab.nio.generator.Generator;
import org.cakelab.nio.meta.FieldMeta;
import org.cakelab.nio.meta.TypeMeta;
import org.cakelab.nio.meta.TypeMeta.Kind;
import org.cakelab.nio.view.DataView;
import org.cakelab.nio.view.array.booleanArrayView;
import org.cakelab.nio.view.array.byteArrayView;
import org.cakelab.nio.view.array.charArrayView;
import org.cakelab.nio.view.array.doubleArrayView;
import org.cakelab.nio.view.array.floatArrayView;
import org.cakelab.nio.view.array.intArrayView;
import org.cakelab.nio.view.array.longArrayView;
import org.cakelab.nio.view.array.shortArrayView;
import org.cakelab.nio.view.scalar.booleanView;
import org.cakelab.nio.view.scalar.byteView;
import org.cakelab.nio.view.scalar.charView;
import org.cakelab.nio.view.scalar.doubleView;
import org.cakelab.nio.view.scalar.floatView;
import org.cakelab.nio.view.scalar.intView;
import org.cakelab.nio.view.scalar.longView;
import org.cakelab.nio.view.scalar.shortView;

public class Utils {
	
	
	public static String generateInitialiser(FieldMeta fieldMeta) {
		
		switch(fieldMeta.getKind()) {
		case Array:
			return generateArrayInitialiser(fieldMeta, 0);
		case Struct:
		case String:
			if (!fieldMeta.isReusable()) throw new Error("trying to create initialiser for non-reusable type in field: " + fieldMeta.getFqN());
			return generateStructInitialiser(fieldMeta, fieldMeta.getTypeMeta(), fieldMeta.view);
		default: 
			throw new Error("TODO");
		}
	}
	
	public static String generateArrayInitialiser(FieldMeta fieldMeta, int startdim) {
		String initialiser = "";
		initialiser = "new " + fieldMeta.getElementType().type.getCanonicalName();
		int maxdim = fieldMeta.getTypeMeta().dimensions;
		if (fieldMeta.isArrayElementReusable()) {
			for (int i = startdim; i < maxdim; i++) {
				initialiser += "[]";
			}
			initialiser += generateArrayComponentsInitialiser(fieldMeta, startdim, maxdim, fieldMeta.lengths);
		} else {
			for (int i = startdim; i < maxdim; i++) {
				initialiser += "[" + fieldMeta.lengths[i] + "]";
			}
			// no array element initialisers possible
		}
		
		return initialiser;
	}

	private static String generateArrayComponentsInitialiser(FieldMeta fieldMeta, int dim, int maxDim, int[] maxLengths) {
		String initialiser = "{";
		for (int i = 0; i < maxLengths[dim]; i++) {
			if (dim+1 < maxDim) {
				initialiser += generateArrayComponentsInitialiser(fieldMeta, dim+1, maxDim, maxLengths);
			} else {
				// This method will be called only, if the elements are reusable structs. 
				assert(fieldMeta.isArrayElementReusable());
				initialiser += generateStructInitialiser(fieldMeta, fieldMeta.getElementType(), fieldMeta.view);
			}
			if (i+1 < maxLengths[dim]) {
				initialiser +=", ";
			}
		}
		initialiser += "}";
		return initialiser;
	}

	public static String generateStructInitialiser(FieldMeta fieldMeta, TypeMeta type, DataView view) {
		if (view != null) {
			return "this." + fieldMeta.name + "." + FactoryMethodGen.METHOD_NAME + "()";
		} else if (type.kind == Kind.Struct) {
			return "this." + fieldMeta.name + "." + FactoryMethodGen.METHOD_NAME + "()";
		} else {
			throw new Error("initialisation method for field " + fieldMeta.getFqN() + "' unknown");
		}
	}


	public static boolean isCompoundField(FieldMeta fm) {
		Kind kind = fm.getKind();
		return kind == Kind.Array || kind == Kind.Struct || kind == Kind.Union;
	}

	public static String generateViewInitialiser(FieldMeta fieldMeta) {
		if (fieldMeta.view == null) {
			Class<?> type = fieldMeta.getKind() == Kind.Array ? fieldMeta.getElementType().type : fieldMeta.getType();
			String viewTypeName = getViewClassFqN(type);
			return "new " + viewTypeName + "((short)" + fieldMeta.align + ")";
		} else {
			String viewTypeName = fieldMeta.view.getClass().getCanonicalName();
			Constructor<? extends DataView> c = fieldMeta.getAppropriateViewConstructor();
			Parameter[] params = c.getParameters();
			if (params.length == 1) {
				return "new " + viewTypeName + "((short)" + fieldMeta.align + ")";
			} else if (params.length == 2) {
				int firstViewDimension = fieldMeta.getTypeMeta().dimensions;
				if (params[1].getType() == int.class) {
					
					return "new " + viewTypeName + "((short)" + fieldMeta.align + ", " +  fieldMeta.lengths[firstViewDimension] + ")";
				} else if (params[1].getType() == int[].class) {
					return "new " + viewTypeName + "((short)" + fieldMeta.align + ", " +  generateDimensionsInitialiser(fieldMeta.lengths, firstViewDimension) + ")";
				} else {
					throw new Error("TODO: method needs to adapt to new API");
				}
			} else {
				throw new Error("TODO: method needs to adapt to new API");
			}
		}
	}
	
	private static String generateDimensionsInitialiser(int[] dimensions, int startDimension) {
		StringBuffer initialiser = new StringBuffer("new int[]{");

		for (int dim = startDimension; dim < dimensions.length; dim++) {
			initialiser.append(dimensions[dim]);
			if (dim+1 < dimensions.length) initialiser.append(",");
		}
		initialiser.append('}');
		return initialiser.toString();
	}


	/**
	 * Returns the type abbreviation used as prefix in ByteBuffer types (such as FloatBuffer) and 
	 * postfix in ByteBuffer methods (such as getFloat()). Methods for type <b><code>byte</code></b> 
	 * (and boolean, mapped to byte) have no postfix.
	 * @param primitive primitive type
	 * @return abbreviation
	 */
	public static String getScalarTypePostfix(Class<?> primitive) {
		if (primitive == byte.class || primitive == boolean.class) {
			return "";
		} else if (primitive == char.class) {
			return "Char";
		} else if (primitive == short.class) {
			return "Short";
		} else if (primitive == int.class) {
			return "Int";
		} else if (primitive == long.class) {
			return "Long";
		} else if (primitive == float.class) {
			return "Float";
		} else if (primitive == double.class) {
			return "Double";
		} else {
			throw new Error("Unknown type postfix for type " + primitive.getSimpleName());
		}
	}
	
	public static Class<?> getScalarBufferType(Class<?> type) {
		if (type == byte.class || type == boolean.class) {
			return java.nio.ByteBuffer.class;
		} else {
			String className = "java.nio."  + getScalarTypePostfix(type) + "Buffer";
			try {
				return Utils.class.getClassLoader().loadClass(className);
			} catch (ClassNotFoundException e) {
				throw new Error(e);
			}
		}
	}


	
	public static String getViewClassFqN(Class<?> dataType) {
		if (dataType.isPrimitive()) {
			return getPrimitiveViewInterface(dataType).getCanonicalName();
		} else {
			return Generator.getStructViewClassFqN(dataType);
		}
	}

	public static String getViewSimpleClassName(Class<?> dataType) {
		if (dataType.isPrimitive()) {
			return getPrimitiveViewInterface(dataType).getSimpleName();
		} else {
			return Generator.getStructViewSimpleClassName(dataType);
		}
	}
	
	public static String getViewPackageName(Class<?> dataType) {
		if (dataType.isPrimitive()) {
			return getPrimitiveViewInterface(dataType).getPackage().getName();
		} else {
			return Generator.getStructViewPackageName(dataType);
		}
	}

	public static Class<?> getPrimitiveViewInterface(Class<?> primitive) {
		assert (primitive.isPrimitive());
		if (primitive == byte.class) return byteView.class;
		else if (primitive == short.class) return shortView.class;
		else if (primitive == int.class) return intView.class;
		else if (primitive == long.class) return longView.class;
		else if (primitive == float.class) return floatView.class;
		else if (primitive == double.class) return doubleView.class;
		else if (primitive == char.class) return charView.class;
		else if (primitive == boolean.class) return booleanView.class;
		else throw new Error("Unsupported primitive type " + primitive.getCanonicalName());
	}

	public static Class<?> getPrimitiveArrayViewInterface(Class<?> primitive) {
		assert (primitive.isPrimitive()) : "not a primitive type";
		if (primitive == byte.class) return byteArrayView.class;
		else if (primitive == short.class) return shortArrayView.class;
		else if (primitive == int.class) return intArrayView.class;
		else if (primitive == long.class) return longArrayView.class;
		else if (primitive == float.class) return floatArrayView.class;
		else if (primitive == double.class) return doubleArrayView.class;
		else if (primitive == char.class) return charArrayView.class;
		else if (primitive == boolean.class) return booleanArrayView.class;
		else throw new Error("Unsupported primitive type " + primitive.getCanonicalName());
	}

	
	public static String toCamelCase(String name) {
		return Character.toUpperCase(name.charAt(0)) + name.substring(1);
	}



}
