package org.cakelab.nio.generator.codegen.struct;

import java.io.IOException;
import java.nio.ByteBuffer;

import org.cakelab.codegen.ClassGen;
import org.cakelab.codegen.IndentStreamControl;
import org.cakelab.codegen.VariableGen;
import org.cakelab.nio.generator.codegen.Utils;
import org.cakelab.nio.meta.FieldMeta;
import org.cakelab.nio.meta.TypeMeta.Kind;

public abstract class ArrayFieldMethodGen extends FieldMethodGen {

	
	
	protected int numIndexParams;

	public ArrayFieldMethodGen(ClassGen classgen, FieldMeta fieldMeta, String typeName, String methodName, int numIndexParams, VariableGen valueParam) {
		super(classgen, fieldMeta, typeName, methodName, createParameterList(numIndexParams, valueParam));
		this.numIndexParams = numIndexParams;
		if (fieldMeta.getKind() != Kind.Array) {
			throw new Error("for attributes of kind array only");
		}
	}

	
	protected String[] getIndexParamNames() {
		int start = parameters.length - numIndexParams;
		String[] result = new String[numIndexParams];
		for (int i = start, param = 0; i < parameters.length; i++, param++) {
			result[param] = parameters[i].getName();
		}
		return result;
	}

	protected String generateIndexParameterList(String[] indices) {
		String result = "";
		for (String index : indices) {
			result += ", " + index;
		}
		return result;
	}
	
	protected String generateIndexedOffset(String[] indices, int maxdim, int[] maxlengths, int stride) {
		String result = "0";
		//
		// indices  = {i,j}
		// maxlens  = {4,8,16}
		// elemsize = 32
		// offset   = i*(elemsize*maxlens[1]*maxlens[0]) + j*elemsize*maxlens[0] + k*elemsize
		//          = ((i*maxlens[1] + j) * maxlens[0] + k) * elemsize;
		if (indices.length > 0) {
			int dimsize = stride;
			for (int dim = 1; dim < maxdim; dim++) {
				dimsize *= maxlengths[dim];
			}
			result = "(";
			for (int dim = 0; dim < indices.length; dim++) {
				result += indices[dim] + "*" + dimsize;
				if (dim+1 < indices.length) {
					result += " + ";
					dimsize /= maxlengths[dim+1];
				}
			}
			result += ")";
		}
		return result;
	}
	
	
	protected boolean isElementAccess() {
		return fieldMeta.getTypeMeta().dimensions == this.numIndexParams;
	}


	
	protected boolean hasValueParam() {
		return (parameters.length - numIndexParams) > 0;
	}
	
	protected String generateForLoopHead(String var, int end) {
		return "for (int " + var + " = 0; " + var + " < " + end + "; " + var + "++) {";
	}

	public void writeArrayAccess(IndentStreamControl pout, Access access, String arrayVariable, String[] indices) throws IOException {
		
		switch (fieldMeta.getElementType().kind) {
		case Scalar:
			if (fieldMeta.view == null) writeArrayAccessNative(pout, access, arrayVariable, indices);
			else writeArrayAccessView(pout, access, arrayVariable, indices);
			break;
		case Struct:
		case String:
			writeArrayAccessView(pout, access, arrayVariable, indices);
			break;
		case Union:
			throw new Error("TODO");
		case Array:
			throw new Error("internal error: kind of most derived array component type cannot be array");
		case Undefined:
		default:
			throw new Error("undefined or unhandeled kind of array component type");
		}
		
	}


	/**
	 * write code to access (get/set) array components of a struct type
	 * @param pout
	 * @param access
	 * @param arrayVariable
	 * @throws IOException
	 */
	private void writeArrayAccessView(IndentStreamControl pout, Access access, String arrayVariable, String[] indices) throws IOException {
		// declare local view variable
		String viewVar = getElemViewAttribute(fieldMeta);
		
		int stride = fieldMeta.getArrayElementStride();

		// declare local size variable
		String sizeVar = "__nio__stride";
		pout.println("int " + sizeVar + " = " + stride + ";");

		int maxdepth = fieldMeta.getTypeMeta().dimensions;

		// declare local offset variable
		String offsetVar = "__nio__elem__offset";
		pout.println("int " + offsetVar + " = " + StructViewClassGen.ATTRIBUTE_STRUCT_OFFSET + " + " + fieldMeta.offset + ";");
		if (indices.length>0) {
			pout.println(offsetVar + " += " + generateIndexedOffset(indices, maxdepth, fieldMeta.lengths, stride) + ";");
		}
		
		// for loop headers
		String loopVars[] = writeForLoopHeaders(pout, fieldMeta, indices.length, maxdepth);

		// offset view
		pout.println(viewVar + ".offset(" + offsetVar + ");");
		
		String arrayComponentAccessExpression = generateArrayComponentAccess(arrayVariable, loopVars);
		String method = access.getEquivalentViewMethod();
		if (access == Access.READ && !fieldMeta.isArrayElementReusable()) {
			pout.println(arrayComponentAccessExpression + " = " + viewVar + "." + method + "();");
		} else {
			pout.println(viewVar + "." + method + "(" + arrayComponentAccessExpression + ");");
		}
		pout.println(offsetVar + " += " + sizeVar + ";");
		
		
		// end for loops
		writeForLoopCloseBrackets(pout, loopVars.length);
		
	}


	private void writeArrayAccessNative(IndentStreamControl pout, Access access, String arrayVariable, String[] indices) throws IOException {


		// declare local offset variable
		String offsetVar = "__nio__member_offset";
		pout.println("int " + offsetVar + " = " + fieldMeta.offset + ";");
		if (indices.length>0) {
			pout.println("/* calc indexed offset */");
			pout.println(offsetVar + " += " + generateIndexedOffset(indices, fieldMeta.lengths.length, fieldMeta.lengths, fieldMeta.getArrayElementStride()) + ";");
		}
		
		

		boolean multiElemAccess = fieldMeta.getTypeMeta().dimensions > indices.length;
		
		if (multiElemAccess) {
			if (fieldMeta.getArrayElementStride() > fieldMeta.getArrayElementSize()) 
			{
				writeMultiElemAccessNativeSpaced(pout, access, arrayVariable, indices, offsetVar);
			}
			else 
			{
				writeMultiElemAccessNativePacked(pout, access, arrayVariable, indices, offsetVar);
			}
		} else if (fieldMeta.view != null) {
			throw new Error("need to call writeArrayAccessView");
		} else {
			if (access == Access.READ) pout.println(generateScalarGetter(fieldMeta.getElementType().type, offsetVar, arrayVariable) + ";");
			else  pout.println(generateScalarSetter(fieldMeta.getElementType().type, offsetVar, arrayVariable) + ";");
		}
	}


	private void writeMultiElemAccessNativeSpaced(IndentStreamControl pout, Access access, String arrayVariable,
			String[] indices, String offsetVar) throws IOException {
		assert(fieldMeta.getArrayElementStride() > fieldMeta.getArrayElementSize());
		
		Class<?> elemType = fieldMeta.getElementType().type;
		int maxdepth = fieldMeta.getTypeMeta().dimensions;

		String bufferVar = StructViewClassGen.ATTRIBUTE_BUFFER;

		String bufferAccessMethod = access.getEquivalentBufferMethod() + Utils.getScalarTypePostfix(elemType);

		String totalOffset = StructViewClassGen.ATTRIBUTE_STRUCT_OFFSET + " + " + offsetVar;

		
		// loop
		String loopVars[] = writeForLoopHeaders(pout, fieldMeta, indices.length, maxdepth);
		
		String arrayComponentAccessExpression = generateArrayComponentAccess(arrayVariable, loopVars);
		
		if (elemType == boolean.class) {
			if (access == Access.READ) {
				pout.println(arrayComponentAccessExpression + " = " + bufferVar + "." + bufferAccessMethod + "(" + totalOffset + ")==0?false:true;");
			} else {
				pout.println(bufferVar + "." + bufferAccessMethod + "(" + totalOffset + ", (byte)(" + arrayComponentAccessExpression + "?1:0));");
			}
		} else {
			pout.println(bufferVar + "." + bufferAccessMethod + "("  + totalOffset + ", " + arrayComponentAccessExpression + ");");
		}

		pout.println(offsetVar + " += " + fieldMeta.getArrayElementStride() + ";");
		
		writeForLoopCloseBrackets(pout, loopVars.length);
		// endloop
	}


	private void writeMultiElemAccessNativePacked(IndentStreamControl pout, Access access, String arrayVariable,
			String[] indices, String offsetVar) throws IOException {
		int maxdepth;
		Class<?> componentType = fieldMeta.getElementType().type;
		if (componentType == boolean.class) {
			maxdepth = fieldMeta.getTypeMeta().dimensions;
		} else {
			maxdepth = fieldMeta.getTypeMeta().dimensions - 1;
		}
		
		
		pout.println("" + StructViewClassGen.ATTRIBUTE_BUFFER + ".position(" + StructViewClassGen.ATTRIBUTE_STRUCT_OFFSET + " + " + offsetVar + ");");

		String bufferVar;
		
		Class<?> bufferType = Utils.getScalarBufferType(fieldMeta.getElementType().type);
		if (bufferType != ByteBuffer.class) {
			bufferVar = "tmpBuffer";
			pout.println(bufferType.getCanonicalName() + " " + bufferVar + " = " + StructViewClassGen.ATTRIBUTE_BUFFER + ".as" + bufferType.getSimpleName() + "();");
		} else {
			bufferVar = StructViewClassGen.ATTRIBUTE_BUFFER;
		}

		String bufferAccessMethod = access.getEquivalentBufferMethod();

		
		String loopVars[] = writeForLoopHeaders(pout, fieldMeta, indices.length, maxdepth);
		String arrayComponentAccessExpression = generateArrayComponentAccess(arrayVariable, loopVars);
		
		
		if (componentType == boolean.class) {
			if (access == Access.READ) {
				pout.println(arrayComponentAccessExpression + " = " + bufferVar + "." + bufferAccessMethod + "()==0?false:true;");
			} else {
				pout.println(bufferVar + "." + bufferAccessMethod + "((byte)(" + arrayComponentAccessExpression + "?1:0));");
			}
		} else {
			pout.println(bufferVar + "." + bufferAccessMethod + "(" + arrayComponentAccessExpression + ");");
		}
		
		writeForLoopCloseBrackets(pout, loopVars.length);
	}


	/**
	 * @param pout
	 * @param depth depth of the for loop (e.g. the length of the for loop variables received from writeForLoopHeaders
	 * @throws IOException
	 */
	private void writeForLoopCloseBrackets(IndentStreamControl pout, int depth) throws IOException {
		for (int i = 0; i < depth; i++) {
			pout.indent(-1);
			pout.println("}");
		}
	}


	/**
	 * Generates an expression to access a particular component of the 
	 * array using all of the loop variable given.
	 * 
	 * e.g. "a[i][j]"
	 * 
	 * @param loopVars list of loop variables received from writeLoopHeaders
	 * @param arrayVariable the member of a class which is an array or an expression which resolves to an array variable
	 * @param indices 
	 * @return
	 */
	private String generateArrayComponentAccess(String arrayVariable, String[] loopVars) {
		String arrayField = arrayVariable;
		for (int i = 0; i < loopVars.length; i++) {
			arrayField += "[" + loopVars[i] + "]";
		}
		return arrayField;
	}


	/**
	 * Writes the for loop headers to iterate over all components in all dimensions of the array
	 * @param pout
	 * @param fieldMeta field meta of the member attribute, which is the array to be enumerated
	 * @param depth defines the maximum depth to go in the for loop. The length of the list of 
	 *        loop variables will consequently have the same length.
	 * @param maxdepth 
	 * @return list of for loop variables in the order of dimensions of the array
	 * @throws IOException
	 */
	private String[] writeForLoopHeaders(IndentStreamControl pout, FieldMeta fieldMeta, int startdepth, int maxdepth) throws IOException {
		String loopVars[] = new String[maxdepth-startdepth];
		for (int i = startdepth; i < maxdepth; i++) {
			int depth = i - startdepth;
			loopVars[depth] = "index" + i;
			pout.println(generateForLoopHead(loopVars[depth], fieldMeta.lengths[i]));
			pout.indent(+1);
		}
		return loopVars;
	}


	/**
	 * Writes an assertion expression to the given stream pout, which checks 
	 * all dimensions of a given array 'param' to have the the dimension and length
	 * as declared for the corresponding field of the class.
	 * @param pout output stream to receive the output
	 * @param array expression which evaluates to a reference on an array to be asserted
	 * @throws IOException
	 */
	public void writeArrayAssertion(IndentStreamControl pout, String array, String[] indices) throws IOException {
		pout.println("assert(");
		pout.indent(+1);
		writeArrayCheck(pout, array, indices);
		pout.indent(-1);
		pout.println(");");
	}

	/**
	 * writes expression to validate the dimensions and length of an array against its declaration, 
	 * which evaluates to a boolean to be used in an assertion.
	 * @param pout
	 * @param array expression which evaluates to a reference on an array instance 
	 * @throws IOException
	 */
	public void writeArrayCheck(IndentStreamControl pout, String array, String[] indices) throws IOException {
		pout.println(array + " != null");
		String dimAccess = "";
		for (int dim = indices.length; dim < fieldMeta.getTypeMeta().dimensions; dim++) {
			pout.println("&& " + array + dimAccess + ".length == " + fieldMeta.lengths[dim]);
			dimAccess +="[0]";
		}
	}

	
	private static VariableGen[] createParameterList(int numIndexParameters, VariableGen valueParam) {
		int numParams = numIndexParameters;
		VariableGen[] params;
		int i = 0;
		if (valueParam != null) {
			numParams++;
			params = new VariableGen[numParams];
			params[i++] = valueParam;
		} else {
			params = new VariableGen[numParams];
		}
		for (int dim = 0; i < numParams; i++, dim++) {
			params[i] = new VariableGen(int.class, "index" + dim);
		}
		return params;
	}
	
	protected static Class<?> getAccessedComponentType(Class<?> type, int numIndexParams) {
		for (int dim = 0; dim < numIndexParams; dim++) {
			type = type.getComponentType();
			if (type == null) {
				throw new Error("wrong number of indices");
			}
		}
		return type;
	}


}
