package org.cakelab.nio.generator.codegen.struct;

import java.io.IOException;

import org.cakelab.codegen.ClassGen;
import org.cakelab.codegen.IndentStreamControl;
import org.cakelab.codegen.VariableGen;
import org.cakelab.nio.generator.Generator;
import org.cakelab.nio.generator.codegen.Utils;
import org.cakelab.nio.meta.FieldMeta;
import org.cakelab.nio.meta.TypeMeta;
import org.cakelab.nio.meta.TypeMeta.Kind;
import org.cakelab.nio.view.array.CArrayView;
import org.cakelab.nio.view.string.StringView;

public class ArrayFieldViewMethodGen extends ArrayFieldMethodGen {

	public ArrayFieldViewMethodGen(ClassGen classgen, FieldMeta fieldMeta, int numIndexParams, boolean addValueParam) {
		super(classgen, 
				fieldMeta, 
				getAccessedComponentViewType(fieldMeta, numIndexParams), 
				"view" + Utils.toCamelCase(fieldMeta.name), 
				numIndexParams, getResultParam(addValueParam, fieldMeta, numIndexParams)
		);
		if (fieldMeta.getElementType().kind == Kind.Scalar) {
			Class<?> viewClass = Utils.getPrimitiveArrayViewInterface(fieldMeta.getElementType().type);
			classgen.addImport(viewClass.getCanonicalName());
			if (fieldMeta.lengths.length > 1) {
				classgen.addImport(CArrayView.class.getCanonicalName());
			}
		} else {
			classgen.addImport(CArrayView.class.getCanonicalName());
		}
		
	}
	
	public static VariableGen getResultParam(boolean addValueParam, FieldMeta fieldMeta, int numIndexParams) {
		if (addValueParam) return new VariableGen(getAccessedComponentViewType(fieldMeta, numIndexParams), "view");
		else return null;
	}

	public static String getAccessedComponentViewType(FieldMeta fieldMeta, int dimension) {
		StringBuffer typeName = new StringBuffer();
		// template parameter list recursive
		appendAccessedComponentViewType(typeName, getAccessedComponentType(fieldMeta.getType(), dimension), getElemViewType(fieldMeta));
		return typeName.toString();
	}

	private static String getElemViewType(FieldMeta fieldMeta) {
		if (fieldMeta.view != null) return fieldMeta.view.getClass().getCanonicalName();
		else {
			TypeMeta elemTypeMeta = fieldMeta.getElementType();
			Class<?> elemType = elemTypeMeta.type;
			switch(elemTypeMeta.kind) {
			case Array:
				throw new Error("internal error: array " + fieldMeta.name + " can't have an array as element type");
			case String:
				return StringView.class.getSimpleName();
			case Scalar:
				return Utils.getPrimitiveViewInterface(elemType).getSimpleName();
			case Struct:
				return Generator.getStructViewClassFqN(fieldMeta.getElementType().type);
			case Union:
				throw new Error("union not yet implemented");
			case Undefined:
			default:
				throw new Error("internal error: undefined or unexpected type in array " + fieldMeta.name);
			}
		}
	}

	private static void appendAccessedComponentViewType(StringBuffer typeName, Class<?> type, String elemViewType) {
		if (type.isArray()) {
			Class<?> componentType = type.getComponentType();
			if (componentType.isPrimitive()) {
				typeName.append(getPrimitiveArrayViewType(componentType));
			} else {
				// component is an array
				typeName.append(CArrayView.class.getSimpleName()).append('<');
				typeName.append(componentType.getCanonicalName());
				typeName.append(',');
				appendAccessedComponentViewType(typeName, componentType, elemViewType);
				typeName.append('>');
			}
		} else {
			typeName.append(elemViewType);
		}
	}

	private static String getPrimitiveArrayViewType(Class<?> componentType) {
		return Utils.getPrimitiveArrayViewInterface(componentType).getSimpleName();
	}

	
	
	@Override
	public void writeMethodBody(IndentStreamControl pout) throws IOException {

		String[] indices = getIndexParamNames();

		
		if (!hasValueParam()) {
			String viewCreation = generateArrayViewInitialiser(new StringBuffer(), fieldMeta, this.numIndexParams).toString();
			String viewVar = "view";
			pout.println(this.typeName + " " + viewVar + " = " + viewCreation + ";");
			pout.println(viewVar + ".open(" + StructViewClassGen.ATTRIBUTE_BUFFER + ", 0);");
			String call = this.name + "(" + viewVar;
			for (int i = 0; i < parameters.length; i++) {
				call += ", " + parameters[i].getName();
			}
			call += ")";
			pout.println("return " + call + ";");
		} else {
			// declare local offset variable
			String offsetVar = "__nio__member_offset";
			pout.println("int " + offsetVar + " = " + StructViewClassGen.ATTRIBUTE_STRUCT_OFFSET + " + " + fieldMeta.offset + ";");
			if (indices.length>0) {
				pout.println("/* calc indexed offset */");
				pout.println(offsetVar + " += " + generateIndexedOffset(indices, fieldMeta.lengths.length, fieldMeta.lengths, fieldMeta.getArrayElementStride()) + ";");
			}
			
			pout.println("view.offset(" + offsetVar + ");");
			pout.println("return view;");
		}

	}

	private StringBuffer generateArrayViewInitialiser(StringBuffer viewCreation, FieldMeta fieldMeta, int dim) {
		
		if (dim == fieldMeta.getArrayDepth()) {
			viewCreation.append(Utils.generateViewInitialiser(fieldMeta));
		} 
		else if (fieldMeta.getElementType().kind == Kind.Scalar && dim+1 == fieldMeta.getArrayDepth()) 
		{
			String factoryClass = Utils.getPrimitiveArrayViewInterface(fieldMeta.getElementType().type).getSimpleName();
			if (fieldMeta.view != null) {
				viewCreation.append(factoryClass).append(".create(").append(Utils.generateViewInitialiser(fieldMeta)).append(", ").append(fieldMeta.lengths[dim]).append(")");
			} else {
				viewCreation.append(factoryClass).append(".create(").append(fieldMeta.lengths[dim]).append(")");
			}
		} 
		else 
		{
			viewCreation.append("new ");
			appendAccessedComponentViewType(viewCreation, getAccessedComponentType(fieldMeta.getType(), dim), getElemViewType(fieldMeta));
			viewCreation.append("(");
			generateArrayViewInitialiser(viewCreation, fieldMeta, dim+1);
			viewCreation.append(", " + fieldMeta.lengths[dim] + ")");
		}
		return viewCreation;
	}




}
