package org.cakelab.nio.generator.codegen.struct;

import java.lang.reflect.Modifier;
import java.nio.ByteBuffer;

import org.cakelab.codegen.JavaDoc;
import org.cakelab.codegen.SourceCodeUnit;
import org.cakelab.codegen.VariableGen;
import org.cakelab.nio.generator.Generator;
import org.cakelab.nio.generator.codegen.AlignmentConstructorGen;
import org.cakelab.nio.generator.codegen.CloneMethodGen;
import org.cakelab.nio.generator.codegen.CloseMethodGen;
import org.cakelab.nio.generator.codegen.CopyConstructorGen;
import org.cakelab.nio.generator.codegen.DefaultConstructorGen;
import org.cakelab.nio.generator.codegen.FactoryMethodGen;
import org.cakelab.nio.generator.codegen.JavaTypeMethodGen;
import org.cakelab.nio.generator.codegen.OffsetMethodGen;
import org.cakelab.nio.generator.codegen.OpenMethodGen;
import org.cakelab.nio.generator.codegen.SizeofMethodGen;
import org.cakelab.nio.generator.codegen.ViewAttributeGen;
import org.cakelab.nio.generator.codegen.ViewClassGen;
import org.cakelab.nio.meta.FieldMeta;
import org.cakelab.nio.meta.MetadataDictionary;
import org.cakelab.nio.meta.MetadataUtils;
import org.cakelab.nio.meta.TypeMeta;
import org.cakelab.nio.meta.TypeMeta.Kind;
import org.cakelab.nio.view.impl.StructViewBase;

public class StructViewClassGen extends ViewClassGen {
	public static final String ATTRIBUTE_BUFFER = "__nio__buffer";
	public static final String ATTRIBUTE_STRUCT_OFFSET = "__nio__offset";
	public static final String ATTRIBUTE_STRUCT_ALIGNMENT = "__nio__alignment";
	
	private TypeMeta structMeta;
	private Generator generator;

	public StructViewClassGen(Generator generator, SourceCodeUnit sourceCodeUnit, String pkg, String className, Class<?> structType) {
		this(generator, sourceCodeUnit, "public", pkg, className, structType);
	}

	public StructViewClassGen(Generator generator, SourceCodeUnit code, String modifiers, String pkg, String className, Class<?> structType) {
		super(generator, code, modifiers, pkg, className, StructViewBase.class.getSimpleName() + "<" + structType.getCanonicalName() + ">");

		checkStructClass(structType);
		
		this.generator = generator;
		this.structMeta = MetadataDictionary.get(structType);
		addInterface(getViewInterface(structMeta).getCanonicalName(), getInterfaceExpression(structMeta));
		addImport(StructViewBase.class.getCanonicalName());
		setJavaDoc(createJavaDoc());
	}


	private JavaDoc createJavaDoc() {
		JavaDoc doc = new JavaDoc("/**\n");
		doc.appendln("Generated view class for {@link " + structMeta.type.getCanonicalName() + "}");
		return doc;
	}

	private String getInterfaceExpression(TypeMeta structMeta) {
		return getViewInterface(structMeta).getSimpleName() + "<" + structMeta.type.getCanonicalName() + ">";
	}

	public static void checkStructClass(Class<?> structType) {
		
		String context = "in " + structType.getCanonicalName() + ":\n";
		
		int modifiers = structType.getModifiers();
		if (!Modifier.isPublic(modifiers)) throw new Error(context + "struct classes must be public");

		if (structType.isAnonymousClass()) {
			throw new Error(context + "anonymous classes not supported");
		} else if (structType.isInterface()) {
			throw new Error(context + "interfaces not supported");
		} else if (structType.isEnum()) {
			throw new Error(context + "enums not supported");
		} else if (structType.isLocalClass()) {
			throw new Error(context + "local classes not supported");
		} else if (structType.isMemberClass()) {
			if (!Modifier.isStatic(modifiers)) throw new Error(context + "member classes must be static");
		}
		
		if (structType.getSuperclass() != Object.class) {
			throw new Error(context + "struct classes do not support inheritance (super classes)");
		}
	}

	@Override
	public void run() {
		populateBody();
	}

	protected void populateBody() {
		boolean needsOpenCloseOverride = false;

		
		addVersionCheck();
		
		// add member classes to the view
		for (Class<?> memberClass : structMeta.type.getClasses()) {
			int modifiers = memberClass.getModifiers();
			if (Modifier.isPublic(modifiers) && Modifier.isStatic(modifiers)) {
				analyseMemberClass(memberClass);
			}
		}
		

		addMethod(new CopyConstructorGen(this));
		addMethod(new AlignmentConstructorGen(this, structMeta));
		addMethod(new DefaultConstructorGen(this, structMeta));

		
		// generate access methods
		for (FieldMeta f : structMeta.fields) {

			
			FieldMeta fieldMeta = structMeta.getField(f.name);
			
			String viewTypeName = null;
			if (fieldMeta.view != null)	{
				viewTypeName = fieldMeta.view.getClass().getCanonicalName();
			}
			
			switch (fieldMeta.getKind()) {
			case Scalar:
				addMethod(new ScalarFieldGetterMethodGen(this, fieldMeta));
				addMethod(new ScalarFieldSetterMethodGen(this, fieldMeta));
				break;
			case Array:

				if (viewTypeName == null && ArrayFieldViewMethodGen.usesElemView(fieldMeta)) {
					viewTypeName = handleMissingView(fieldMeta.getElementType().type);
				}
				
				int last = fieldMeta.getTypeMeta().dimensions+1;
				for (int dim = 0; dim < last; dim++) {
					boolean addValueParam = dim < (last-1) || fieldMeta.isArrayElementReusable();
					if (addValueParam) {
						addMethod(new ArrayFieldGetterMethodGen(this, fieldMeta, dim, true));
					}
					addMethod(new ArrayFieldGetterMethodGen(this, fieldMeta, dim, false));
					addMethod(new ArrayFieldSetterMethodGen(this, fieldMeta, dim));
				}
				
				addMethod(new ArrayFieldViewMethodGen(this, fieldMeta, 0, false));
				addMethod(new ArrayFieldViewMethodGen(this, fieldMeta, 0, true));
				if (   (last > 1)
					&& !(fieldMeta.getElementType().kind == Kind.Scalar 
					     && fieldMeta.view == null)
				) 
				{
					addMethod(new ArrayFieldViewMethodGen(this, fieldMeta, last-1, false));
					addMethod(new ArrayFieldViewMethodGen(this, fieldMeta, last-1, true));
				}
				break;
			case Struct:
			case String:
				// delegate generation of another view for the type of this field
				if (viewTypeName == null && StructFieldViewMethodGen.usesElemView(fieldMeta)) {
					viewTypeName = handleMissingView(fieldMeta.getType());
				}
				if (fieldMeta.isReusable()) {
					addMethod(new StructFieldGetterMethodGen(this, fieldMeta, new VariableGen(fieldMeta.getType(), "result")));
				}
				addMethod(new StructFieldGetterMethodGen(this, fieldMeta));
				addMethod(new StructFieldSetterMethodGen(this, fieldMeta));
				addMethod(new StructFieldViewMethodGen(this, viewTypeName, fieldMeta));
				addMethod(new StructFieldViewMethodGen(this, viewTypeName, fieldMeta, new VariableGen(viewTypeName, "result")));
				break;
			case Undefined:
				throw new Error("undefined kind of type");
			default:
				throw new Error("TODO");
			}
			
			if (viewTypeName != null) {
				addField(new ViewAttributeGen(viewTypeName, fieldMeta));
				needsOpenCloseOverride = true;
			}

		}

		
		
		// generate implementations of standard methods declared in interface StructView
		if (needsOpenCloseOverride) {
			addMethod(new OpenMethodGen(this, structMeta, new VariableGen(ByteBuffer.class, "buffer"), new VariableGen(int.class, "offset")));
			addMethod(new OffsetMethodGen(this, structMeta));
			addMethod(new CloseMethodGen(this, structMeta));
		}
		addMethod(new StructGetterMethodGen(this, 0, structMeta));
		addMethod(new StructGetterMethodGen(this, 0, structMeta, new VariableGen(structMeta.type, "result")));
		addMethod(new StructSetterMethodGen(this, 0, structMeta));
		addMethod(new CloneMethodGen(this));
		addMethod(new CloneMethodGen(this, new VariableGen(getClassName(), "view")));
		addMethod(new JavaTypeMethodGen(this, structMeta));
		addMethod(new SizeofMethodGen(this, structMeta, "sizeof"));
		addMethod(new FactoryMethodGen(this, structMeta));
	}

	private void addVersionCheck() {
		setStaticBlock(new StaticVersionCheck());
	}

	private String handleMissingView(Class<?> type) {
		assert (type != null);
		// schedule generation of that view if not member of this class 
		Class<?> enclosingClass = MetadataUtils.getToplevelEnclosingClass(type);
		if (enclosingClass != this.structMeta.type) {
			generator.scheduleDataViewGeneration(enclosingClass);
		}
		
		if (!Generator.getStructViewPackageName(type).equals(this.pkgName)) {
			addImport(Generator.getStructViewClassFqN(type));
		}
		return Generator.getStructViewSimpleClassName(type);
	}

	private void analyseMemberClass(Class<?> memberClass) {
		String fqn = Generator.getStructViewClassFqN(memberClass);
		if (!memberClasses.containsKey(fqn)) {
			StructViewClassGen classGen = new StructViewClassGen(generator, sourceCodeUnit, "public static", Generator.getStructViewPackageName(memberClass), Generator.getStructViewSimpleClassName(memberClass), memberClass);
			classGen.populateBody();
			memberClasses.put(fqn, classGen);
		}
	}


}
