package org.cakelab.nio.generator.codegen.struct;

import org.cakelab.codegen.ClassGen;
import org.cakelab.codegen.IndentStreamControl;
import org.cakelab.codegen.MethodGen;
import org.cakelab.codegen.VariableGen;
import org.cakelab.nio.generator.codegen.Utils;
import org.cakelab.nio.meta.FieldMeta;
import org.cakelab.nio.meta.TypeMeta.Kind;

public abstract class ViewClassMethodGen extends MethodGen {

	public ViewClassMethodGen(ClassGen classgen, String modifiers, Class<?> type, String methodName, VariableGen ... parameters) {
		this(classgen, modifiers, type.getCanonicalName(), methodName, parameters);
	}
	
	public ViewClassMethodGen(ClassGen classgen, String modifiers, String type, String methodName, VariableGen ... parameters) {
		super(classgen, modifiers, type, methodName, parameters);
	}
	
	protected void writeSaveOffset(IndentStreamControl pout) {
		pout.println("int __nio__current_offset = " + StructViewClassGen.ATTRIBUTE_BUFFER + ".position();");
	}

	protected void writeResetOffset(IndentStreamControl pout) {
		pout.println("" + StructViewClassGen.ATTRIBUTE_BUFFER + ".position(__nio__current_offset);");
	}

	protected String generateFieldGetter(FieldMeta fm) {
		return generateFieldGetter(fm, null);
	}

	protected String generateFieldSetter(FieldMeta fm, String tmpVariable) {
		switch(fm.getKind()) {
		case Scalar: 
			if (fm.view == null) return generateScalarSetter(fm, tmpVariable);
			else return generateStructSetter(fm, tmpVariable);
		case Array: 
			return generateArraySetter(fm, tmpVariable);
		case Struct: 
		case String:
			return generateStructSetter(fm, tmpVariable);
		case Union:
			throw new Error("TODO");
		case Undefined:
		default:
			throw new Error("Undefined kind of type");
		}
	}

	protected String generateFieldGetter(FieldMeta fm, String resultVar) {
		switch(fm.getKind()) {
		case Scalar:
			if (fm.view == null) return generateScalarGetter(fm, resultVar);
			else return generateStructGetter(fm, resultVar);
		case Array: 
			return generateArrayGetter(fm, resultVar);
		case Struct: 
		case String:
			return generateStructGetter(fm, resultVar);
		case Union:
			throw new Error("TODO");
		case Undefined:
		default:
			throw new Error("Undefined kind of type");
		}
	}
	
	protected String generateStructGetter(FieldMeta fm, String resultVar) {
		if (fm.isReusable()) {
			return "get" + Utils.toCamelCase(fm.name) + "(" + resultVar + ")";
		} else {
			String assignment = resultVar != null ? resultVar + " = " : "";
			if (fm.view == null) {
				// without view, delegate to get<name>()
				return assignment +  "get" + Utils.toCamelCase(fm.name) + "()";
			}
			else 
			{
				// with view, use the view instead
				return assignment + getElemViewAttribute(fm) + ".get()";
			}
		}
	}


	protected String generateStructSetter(FieldMeta fm, String valueExpression) {
		if (fm.view == null) {
			return "set" + Utils.toCamelCase(fm.name) + "(" + valueExpression + ")";
		} else {
			return getElemViewAttribute(fm) + ".set(" + valueExpression + ")";
		}
	}

	protected String getElemViewAttribute(FieldMeta fm) {
		return "this." + fm.name;
	}

	public static boolean usesElemView(FieldMeta f) {
		if (f.view != null) return true;
		else {
			switch(f.getKind()) {
			case Array:
				return f.getElementType().kind != Kind.Scalar;
			case Scalar:
				return false;
			case String:
			case Struct:
				return true;
			case Undefined:
			case Union:
			default:
				throw new Error("undefined or unhandeled kind of type");
			}
		}
	}


	
	protected String generateArrayGetter(FieldMeta fm, String resultVar) {
		return "get" + Utils.toCamelCase(fm.name) + "(" + resultVar + ")";
	}

	protected String generateArraySetter(FieldMeta fm, String valueExpression) {
		return "set" + Utils.toCamelCase(fm.name) + "(" + valueExpression + ")";
	}

	protected String generateScalarGetter(FieldMeta fm, String resultVar) {
		return generateScalarGetter(fm.getType(), Integer.toString(fm.offset), resultVar);
	}

	protected String generateScalarGetter(Class<?> type, String offset, String resultVar) {
		String expression = "";
		String nativeGet;
		
		offset = StructViewClassGen.ATTRIBUTE_STRUCT_OFFSET + " + " + offset;
		
		nativeGet = StructViewClassGen.ATTRIBUTE_BUFFER + ".get" + Utils.getScalarTypePostfix(type) + "(" + offset + ")";
		
		if (type == boolean.class) nativeGet = nativeGet + "==0?false:true";
		if (resultVar != null) {
			expression += resultVar + " = " + nativeGet;
		} else {
			expression += nativeGet;
		}
		return expression;
	}

	/**
	 * @param type type of the data to be set in the buffer
	 * @param offset offset of the data
	 * @param expression expression which evaluates to the value of the data to be set (e.g. variable name)
	 * @return access expression (e.g. buffer.put(0, value) )
	 */
	protected String generateScalarSetter(FieldMeta fm, String expression) {
		return generateScalarSetter(fm.getType(), Integer.toString(fm.offset), expression);
	}

	protected String generateScalarSetter(Class<?> type, String offset, String expression) {
		if (type == boolean.class) expression = "(byte)(" + expression + "?1:0)";
		offset = StructViewClassGen.ATTRIBUTE_STRUCT_OFFSET + " + " + offset;
		return "" + StructViewClassGen.ATTRIBUTE_BUFFER + ".put" + Utils.getScalarTypePostfix(type) + "(" + offset + ", " + expression + ")";
	}

}
