package org.cakelab.nio.meta;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Modifier;
import java.lang.reflect.Parameter;
import java.util.Arrays;

import org.cakelab.nio.meta.TypeMeta.Kind;
import org.cakelab.nio.view.DataView;
import org.cakelab.nio.view.annotations.nio;
import org.cakelab.nio.view.string.StringView;

/**
 * Represents meta data extracted from a field of
 * a class which is meant to represent the declaration 
 * of a struct in the context of Java Native I/O.
 * 
 * @author homac
 *
 */
public class FieldMeta {


	private final Field field;
	private final TypeMeta typemeta;
	/** Name of the field */
	public final String name;
	/** Offset of that field relative to the start of the struct. */
	public final int offset;
	/** Effective size of this field in native data representation. 
	 * Note: the size of an array for example includes the size of 
	 * all of its components and their alignment. 
	 * @see #lengths*/
	public final int size;
	
	/** Alignment of this field in respect to the declared alignment
	 * and the alignment derived from nested types.
	 */
	public final int align;

	/** Whether objects of the associated type of this field are supposed to be reusable (e.g. not immutable). */
	private final boolean reusable;
	
	/** Declared for array or sequence types only!
	 * <p>
	 *   The length of each dimension of the array or sequence type.
	 * </p>. */
	public final int[] lengths;
	/**
	 * View type (actual an instance) associated with this field.
	 * <p>
	 * This is either the view, that was given by the annotation @nio(view=MyViewClass) or
	 * a standard view if needed (e.g. for strings). */
	public final DataView view;
	
	public FieldMeta(Field f, nio typeAnnotation, TypeMeta typemeta, int offset, boolean analyse) {
		this.field = f;
		this.name = f.getName();
		this.typemeta = typemeta;
		this.offset = offset;
		nio fieldAnnotation = f.getAnnotation(nio.class);

		int size = 0;
		
		lengths = evaluateLengths(fieldAnnotation);
		align = evaluateAlign(fieldAnnotation, typemeta.align);

		
		switch(typemeta.kind) {
		case Array:
			if (lengths == null) {
				error("arrays must have a field annotation declaring its length for each dimension (e.g. @nio(length={2,4})  int[][] arr;)");
			}
			if (typemeta.dimensions > lengths.length) {
				error("arrays must declare its length for every dimension (e.g. @nio(length={2,4})  int[][] arr;)");
			} 
			
			view = evaluateView(fieldAnnotation);
			if (typemeta.dimensions < lengths.length && view == null) {
				error("regular arrays must have the same number of dimensions as declared in the 'length' attribute of its 'nio' annotation");
			}
			
			size = lengths[0];
			check_length(lengths, 0);
			for (int dim = 1; dim < typemeta.dimensions; dim++) {
				check_length(lengths, dim);
				size *= lengths[dim];
			}
			size *= getArrayElementStride();
			
			size -= getArrayElementStride() - getArrayElementSize();
			
			
			reusable = evaluateReusable(fieldAnnotation, typemeta.elementTypeMeta.reusable);
			break;
		case String:
			
			if (lengths == null) {
				error("strings must have a field annotation declaring its length (e.g. @nio(length=15)");
			}
			if (lengths.length != 1 || lengths[0] <= 0) {
				error("Strings must have exactly one dimension with a length > 0");
			}
			
			view = evaluateView(fieldAnnotation);
			size = view.sizeof();
			reusable = false;
			break;
		case Struct:
			view = evaluateView(fieldAnnotation);
			size = (view != null) ? view.sizeof() : typemeta.size;
			reusable = evaluateReusable(fieldAnnotation, typemeta.reusable);
			break;
		case Scalar:
			view = evaluateView(fieldAnnotation);
			size = (view != null) ? view.sizeof() : typemeta.size;
			reusable = false;
			break;
		default:
			error("undefined or unhandled type");
			reusable = true;
			view = null;
		}
		this.size = size;
	}

	private int[] evaluateLengths(nio fieldAnnotation) {
		return fieldAnnotation != null && fieldAnnotation.length()[0] != nio.LENGTH_NOT_SET ? fieldAnnotation.length() : null;
	}

	DataView evaluateView(nio fieldAnnotation) {
		DataView viewInstance = null;
		Class<? extends DataView> viewClass = null;
		if (fieldAnnotation != null) {
			viewClass = fieldAnnotation.view() != nio.VIEW_NOT_SET ? fieldAnnotation.view() : null;
		}
		if (viewClass == null) {
			viewClass = typemeta.view;
		}
		if (viewClass == null) {
			viewClass = getStandardViewClass();
		}
		if (viewClass != null) {
			Constructor<? extends DataView> c = null;
			try {

				c = getAppropriateViewConstructor(viewClass);
				viewInstance = createViewInstance(c);
				
			} catch (InstantiationException | IllegalAccessException | IllegalArgumentException
					| InvocationTargetException e) {
				error("error instantiating view " + viewClass.getCanonicalName(), e);
			} catch (NullPointerException e) {
				error("internal error: cannot instantiate view of type '" + viewClass.getCanonicalName() + "'", e);
			}
		}
		return viewInstance;
	}
	
	private DataView createViewInstance(Constructor<? extends DataView> c) throws InstantiationException, IllegalAccessException, IllegalArgumentException, InvocationTargetException {
		Parameter[] params = c.getParameters();
		assert 0 < params.length && params.length <= 2;
		assert (params[0].getType() == short.class);
		if (params.length == 1) {
			return c.newInstance((short)align);
		} else if (params[1].getType() == int.class) {
			return c.newInstance((short)align, getRemainingDimensions()[0]);
		} else if (params[1].getType() == int[].class) {
			return c.newInstance((short)align, getRemainingDimensions());
		}
		return null;
	}

	public Constructor<? extends DataView> getAppropriateViewConstructor() {
		return view != null ? getAppropriateViewConstructor(view.getClass()) : null;
	}
	
	private Constructor<? extends DataView> getAppropriateViewConstructor(Class<? extends DataView> viewClass) 
	{
		Constructor<? extends DataView> c = null;
		if (hasMoreDimensions()) {
			if ((lengths.length - typemeta.dimensions) == 1) {
				try {
					// trying least expensive first
					c = viewClass.getConstructor(short.class, int.class);
				} catch (Throwable t) {
					// intentionally empty
				}
			}
			if (c == null) {
				try {
					c = viewClass.getConstructor(short.class, int[].class);
				} catch (Throwable t) {
					error("Couldn't find appropriat constructor." + viewClass.getSimpleName() + "(short, int) or " + viewClass.getSimpleName() + "(short, int[])", t);
				}
			}
		}
		else 
		{
			try {
				c = viewClass.getConstructor(short.class);
			} catch (Throwable t) {
				error("Couldn't find appropriat constructor." + viewClass.getSimpleName() + "(short)", t);
			}
		}
		if (!Modifier.isPublic(c.getModifiers())) throw new Error("constructor " + c.toGenericString() + " is supposed to be public");
		return c;
	}

	private Class <? extends DataView> getStandardViewClass() {
		Class<?> type;
		if (typemeta.kind == Kind.Array) {
			type = typemeta.elementTypeMeta.type;
		} else {
			type = typemeta.type;
		}
		if (type == String.class) return StringView.class;
		return null;
	}

	private int[] getRemainingDimensions() {
		return Arrays.copyOfRange(lengths, typemeta.dimensions, lengths.length);
	}

	public boolean hasMoreDimensions() {
		return lengths != null && lengths.length > typemeta.dimensions;
	}

	private boolean evaluateReusable(nio fieldAnnotation, boolean defaultValue) {
		if (fieldAnnotation != null) {
			if (!fieldAnnotation.reusable()) return false;
		}
		return defaultValue;
	}

	private void check_length(int[] lengths, int dim) {
		if (lengths[dim] <= 0) error("length must not be less or equal to 0");
	}

	private int evaluateAlign(nio fieldAnnotation, int typealign) {
		if (fieldAnnotation != null) {
			return Math.max(typealign, fieldAnnotation.align());
		} else {
			return typealign;
		}
	}

	/** returns the type associated with this field */
	public Class<?> getType() {
		return typemeta.type;
	}

	/** returns the kind of the type associated with this field */
	public Kind getKind() {
		return typemeta.kind;
	}

	/** Most derived component type of array. This is the type without the array specification, e.g. for int[][] it is an int.*/
	public TypeMeta getElementType() {
		return typemeta.elementTypeMeta;
	}
	
	private void error(String msg) {
		error(msg, null);
	}

	private void error(String msg, Throwable e) {
		MetadataDictionary.error("in declaration of " + getFqN() + ": " + msg, e);
	}

	/** returns a unique, user readable identifier for this field */
	public String getFqN() {
		return field.toString();
	}

	public TypeMeta getTypeMeta() {
		return typemeta;
	}

	/** returns the true number of dimensions of an array type */
	public int getArrayDepth() {
		Class<?> type = this.getType();
		int depth = 0;
		while(type.isArray()) {
			depth++;
			type = type.getComponentType();
		}
		return depth;
	}

	/** Returns the distance between the starts of two elementary items of an array. */
	public int getArrayElementStride() {
		assert (typemeta.kind == Kind.Array);
		final int MASK = align - 1;
		return getArrayElementSize() + MASK & ~MASK;
	}

	/** returns the size of an elementary item of the array type of this field */
	public int getArrayElementSize() {
		assert (typemeta.kind == Kind.Array);
		if (view != null) return view.sizeof();
		else return typemeta.elementTypeMeta.size;
	}


	/**
	 * defined for arrays only. Determines, whether the 
	 * elementary items of the array are declared to be reusable.
	 * @return true, if elements are reusable
	 */
	public boolean isArrayElementReusable() {
		assert (typemeta.kind == Kind.Array);
		return reusable;
	}
	
	public boolean isReusable() {
		return reusable || typemeta.kind == Kind.Array;
	}


}
