package org.cakelab.nio.meta;

import java.util.HashMap;

import org.cakelab.nio.common.Global;

/**
 * This is a dictionary of meta information on types 
 * in respect to their native representation in the
 * context of Java Native I/O.
 * <p>
 * The dictionary controls parsing of unknown types 
 * (see {@link #get(Class)}) and caches results 
 * internally. To reset the cache call {@link #clear()}.
 * </p>
 * 
 * @author homac
 *
 */
public class MetadataDictionary {
	
	static {
		Global.init();
	}


	
	// The following flags control behaviour
	// when parsing meta data of types.
	// This is used to temporarily suppress warnings 
	// and errors when parsing for proxies (see TypeProxy).
	
	/** no warnings during parsing */
	static int FLAG_NO_WARNING = 1<<0;
	/** ignore errors when parsing */
	static int FLAG_NO_ERROR = 1<<1;
	
	/** 
	 * Controls error and warning notifications. 
	 * All warnings and errors notifications are 
	 * on by default.
	 * @see #FLAG_NO_WARNING
	 * @see #FLAG_NO_ERROR
	 */
	private static int FLAGS = 0;
	
	static boolean ISSET(int flag) {
		return 0 != (FLAGS & flag);
	}
	static void SET(int flag) {
		FLAGS = FLAGS | flag;
	}
	static void UNSET(int flag) {
		FLAGS = FLAGS & ~flag;
	}

	/**
	 * The reference cycle indicator is inserted in the cache for a given type
	 * when the type is currently getting analysed.
	 * When the metadata dictionary gets another request for the same type,
	 * this indicates that there is a cycle in the type dependencies 
	 * (e.g. a type has a member of its own type).
	 */
	private static final TypeMeta REFERENCE_CYCLE_INDICATOR = new TypeMeta();

	/**
	 * This hashmap caches all types seen during its lifetime (e.g. a run of the generator).
	 * This includes all not yet analysed types which are represented by proxies and
	 * the {@link #REFERENCE_CYCLE_INDICATOR}s for types currently getting analysed.
	 */
	static HashMap<Class<?>, TypeMeta> cache = new HashMap<Class<?>, TypeMeta>();
	
	/**
	 * This method returns the meta data for a given type.
	 * 
	 * 
	 * @param dataType Any Java type.
	 * @return Meta data for the given type.
	 */
	public static TypeMeta get(Class<?> dataType) {
		
		TypeMeta result = cache.get(dataType);
		if (result == null || result instanceof TypeProxy) {
			result = getTypeMeta(dataType);
		}
		return result;
	}


	/**
	 * Resets (clears) the meta data cache.
	 */
	public static void clear() {
		cache.clear();
	}

	/**
	 * This method is used when a type needs to be analysed
	 * @param type
	 * @return
	 */
	static TypeMeta getTypeMeta(Class<?> type) {
		TypeMeta result = cache.get(type);
		if (result == null || result instanceof TypeProxy) {
			cache.put(type, REFERENCE_CYCLE_INDICATOR);
			result = new TypeMeta(type);
			cache.put(type, result);
		} else if (result == REFERENCE_CYCLE_INDICATOR) {
			error("detected type reference cycle in use of " + type.getCanonicalName());
		}
		return result;
	}

	/**
	 * 
	 * @param type
	 * @param feedback whether to give warning messages or not.
	 * @return meta data (maybe a proxy)
	 */
	static TypeMeta getTypeMeta(Class<?> type, boolean feedback) {
		if (feedback) return getTypeMeta(type);
		else return getTypeProxy(type);
	}

	
	static TypeMeta getTypeProxy(Class<?> type) {
		int safe = FLAGS;
		try {
			TypeMeta tm = cache.get(type);
			if (tm != null && tm != REFERENCE_CYCLE_INDICATOR) 
			{
				return tm;
			}
			else 
			{
				// temporarily suppress warnings
				SET(FLAG_NO_WARNING);
				SET(FLAG_NO_ERROR);
				TypeProxy typeMeta = new TypeProxy(type);
				if (tm == null) {
					// if there is no type meta yet, then use the proxy until it is analysed
					cache.put(type, typeMeta);
				}
				return typeMeta;
			}
		} finally {
			// reset flags
			FLAGS = safe;
		}
	}


	static void error(String msg) {
		error(msg, null);
	}
	
	static void error(String msg, Throwable e) {
		if (!ISSET(FLAG_NO_ERROR)) {
			if (e == null) throw new Error(msg);
			else throw new Error(msg, e);
		}
	}

	static void warning(String msg) {
		if (!ISSET(FLAG_NO_WARNING)) Global.warning(msg);
	}

}
