package org.cakelab.nio.meta;

import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.HashMap;

import org.cakelab.nio.view.DataView;
import org.cakelab.nio.view.annotations.nio;

/**
 * Represents metadata for a given Java type in 
 * respect to its native representation in the 
 * context of Java Native I/O.
 * 
 * @author homac
 *
 */
public class TypeMeta {
	/** Kind of type which is one of
	 * <ul>
	 * <li>{@link Kind#Scalar}</li>
	 * <li>{@link Kind#Struct}</li>
	 * <li>{@link Kind#Array}</li>
	 * <li>{@link Kind#String}</li>
	 * <li>{@link Kind#Union} (not yet supported)</li>
	 * <li>{@link Kind#Undefined}</li>
	 * <li>{@link Kind#String}</li>
	 * </ul>
	 */
	public enum Kind {
		/** scalar types such as <code>int</code> and <code>float</code> */
		Scalar,
		/** structured types (aka complex types or simply classes). */ 
		Struct,
		/** array types. */
		Array,
		/** Special kind of type java.lang.String. The internal format of a 
		 * string (encoding) is defined by the associated view. 
		 * The standard view for strings is {@link org.cakelab.nio.view.string.StringView}. */
		String,
		/** A structured type, which defines a set of members of different types, which 
		 * will share the same region in memory when the union is instantiated. 
		 * <b>not yet supported!</b> */
		Union,
		/** Value to mark a kind being undefined. */
		Undefined;
	}

	public ArrayList<FieldMeta> fields = new ArrayList<FieldMeta>();
	HashMap<String, FieldMeta> fieldsMap = new HashMap<String, FieldMeta>();
	/** Kind of type. 
	 * @see TypeMeta.Kind */
	public Kind kind = Kind.Undefined;
	/** Associated Java type the meta data was received from. */
	public Class<?> type;
	/** Size of this type in byte.<br/><br/>
	 * <b>Size is undefined for array types!</b><br/> 
	 * The actual size of an array is defined in the {@link FieldMeta} 
	 * using this type.
	 * <br/><br/>
	 * Generally, size is calculated from native size of Java scalar 
	 * types and summed accordingly for arrays and struct types. 
	 * It considers declared alignment and padding. If a view 
	 * was declared for this type, the size is received through method
	 * {@link DataView#sizeof()} of that view instead of actual calculation.
	 * @see FieldMeta#size */
	public int size;
	/** Alignment declared for the type or 1 by default. 
	 */
	public int align;
	/** whether objects of this type are reusable (for example not immutable). */
	public boolean reusable = true;
	// whether this type was fully analysed or acts as a placeholder
	// This will only be false while the parser is running. 
	// Metadata returned to user will always contain parsed information.
	private boolean feedbackGiven;

	
	/* *************** */
	/* FOR ARRAYS ONLY */
	/* *************** */
	
	/** Defined for arrays only! 
	 * This is the type meta of the array elements (not array components!). */
	public TypeMeta elementTypeMeta;
	/** Defined for arrays only! 
	 * This is the number of dimensions of the array. */
	public int dimensions = 0;
	// TODO: review view concept in regards to TypeMeta and standard views.
	public Class<? extends DataView> view;
	
	/** 
	 * This constructor is used to indicate an invalid entry in metadata cache!
	 */
	TypeMeta(){
		// intentionally empty
	}
	
	public TypeMeta(Class<?> type, boolean giveFeedback) {
		
		this.type = type;
		size = 0;
		align = 1;

		Class<?> annotatedType = type.isArray() ? MetadataUtils.getArrayElementType(type) : type;
		nio annotation = annotatedType.getAnnotation(nio.class);
		if (annotation != null) {
			// alignment must be >= 1, thus ...
			align = Math.max(1, annotation.align());
			reusable = annotation.reusable();
			view = annotation.view() != nio.VIEW_NOT_SET ? annotation.view() : null;
		}

		setKind(type);
	
		switch(kind) {
		case Struct:
			Field declaredFields[] = type.getDeclaredFields();
			// Add all accepted fields and determine its alignment.
			// 
			// Alignment of the struct depends on the largest 
			// alignment of any of its fields. Field alignment
			// effects field position in the struct and thereby 
			// struct size but the struct alignment neither affects 
			// field alignment nor struct size.
			for (Field f : declaredFields) {
				if (passFieldFilter(f)) {
					FieldMeta mf = addField(f, giveFeedback);
					align = Math.max(align, mf.align);
				}
			}
			
			if (MetadataUtils.isKnownImmutableClass(type)) {
				reusable = false;
			}
			
			break;
		case Scalar:
			size = MetadataUtils.getScalarSize(type);
			reusable = false;
			break;
		case String:
			reusable = false;
			break;
		case Array:
			/* arrays have at least one dimension */
			dimensions = MetadataUtils.getArrayDimensions(type);
			Class<?> elemType = MetadataUtils.getArrayElementType(type);
			elementTypeMeta = MetadataDictionary.getTypeMeta(elemType, giveFeedback);
			align = Math.max(align, elementTypeMeta.align);
			reusable = reusable && elementTypeMeta.reusable;
			break;
		case Undefined:
			throw new Error("kind of type " + type.getSimpleName() + " undefined");
		default:
			throw new Error("kind of type " + type.getSimpleName() + " not handled");
		}
		feedbackGiven = giveFeedback;
	}


	
	public TypeMeta(Class<?> type) {
		this(type, true);
	}
	
	/**
	 * copy constructor
	 * @param tm
	 */
	public TypeMeta(TypeMeta tm) {
		this.fields = tm.fields;
		this.fieldsMap = tm.fieldsMap;
		this.size = tm.size;
		this.align = tm.align;
		this.type = tm.type;
		this.kind = tm.kind;
		this.reusable = tm.reusable;
		this.elementTypeMeta = tm.elementTypeMeta;
		this.dimensions = tm.dimensions;
		this.view = tm.view;
		this.feedbackGiven = tm.feedbackGiven;
	}

	private boolean passFieldFilter(Field f) {
		if (f.isSynthetic()) return false;
		
		
		int modifiers = f.getModifiers();
		if (Modifier.isStatic(modifiers)) return false;
		if (Modifier.isTransient(modifiers)) return false;
		
		if (!Modifier.isPublic(modifiers)) {
			warning("ignoring invisible field " + f);
			return false;
		}
		return true;
	}

	private void setKind(Class<?> type) {
		// kind may have been set in annotation
		if (kind != Kind.Undefined) return;
		
		if (type.isPrimitive()) {
			kind = Kind.Scalar;
		} else if (type.equals(String.class)) {
			kind = Kind.String;
		} else if (type.isArray()) {
			kind = Kind.Array;
		} else {
			// TODO support Unions
			kind = Kind.Struct;
		}
	}

	private FieldMeta addField(Field f, boolean analyse) {
		nio fieldAnnotation = f.getAnnotation(nio.class);
		boolean hasView = fieldAnnotation != null && fieldAnnotation.view() != null;
		// Generally, fields with a view don't need their types to be analysed.
		// If this field is a field of a type which has not been analysed, then we 
		// don't need to analyse the type of this field either.
		boolean analyseField = (!hasView) && analyse;
		TypeMeta typemeta = MetadataDictionary.getTypeMeta(f.getType(), analyseField);
		int offset = alignOffset(fieldAnnotation, typemeta.align, size);
		FieldMeta mf = new FieldMeta(f, fieldAnnotation, typemeta, offset, analyseField);
		fields.add(mf);
		fieldsMap.put(mf.name, mf);
		size = offset + mf.size;
		return mf;
	}


	private int alignOffset(nio annotation, int align, int offset) {
		if (annotation != null) align = annotation.align();
		int MASK = align-1;
		return (offset + MASK) & ~MASK;
	}

	public FieldMeta getField(String attrName) {
		return fieldsMap.get(attrName);
	}

	private void warning(String msg) {
		MetadataDictionary.warning("in declaration of " + type.getCanonicalName() + ": " + msg);
	}

}
