package org.cakelab.nio.view;

import java.io.Closeable;
import java.nio.ByteBuffer;
import org.cakelab.nio.view.scalar.ScalarView;
import org.cakelab.nio.view.scalar.intView;
import org.cakelab.nio.view.array.CArrayView;
import org.cakelab.nio.view.array.intArrayView;

/** 
 * Base interface for all kinds of data views.
 * 
 * <p>
 * This is the very basic interface for data views. There are 
 * different specific interfaces derived for each major kind of types:
 * </p>
 * <ul>
 * <li>{@link ScalarView}: used by all interfaces for primitive types such as {@link intView}.</li>
 * <li>{@link GenericDataView} is used by all compound types such as classes and arrays.</li>
 * <li>{@link ReusableDataView} is used by all compound types which are not immutable and not primitive types.</li>
 * <li>{@link ArrayView} used by all array interfaces such as {@link CArrayView} or {@link intArrayView}.</li>
 * </ul>
 * @author homac
 *
 */
public interface DataView extends Closeable {
	
	/** 
	 * Open a buffer at its current position (see {@link ByteBuffer#position()}).
	 * 
	 * if you need to consider alignment, then you should call:
	 * <pre>
	 *  view.open(buffer, view.align(buffer.position())
	 * </pre>
	 * <h3>NOTE:</h3> You need to call {@link #close()} when done with buffer access. 
	 * 
	 * 
	 * @param buffer
	 * @return offset from buffer.position()
	 * @see #close()
	 */
	int open(ByteBuffer buffer);
	
	/**
	 * Open a buffer at the given offset. Offset is meant to be a absolute 
	 * offset into the buffer.
	 * Calling this method is the same as
	 * <pre>
	 * MyStructView view = new MyStructView();
	 * buffer.position(offset);
	 * view.open(buffer);
	 * </pre>
	 * <h3>NOTE:</h3> You need to call {@link #close()} when done with buffer access. 
	 * 
	 * @param buffer buffer to be opened
	 * @param offset offset to be used
	 * @see #close()
	 */
	void open(ByteBuffer buffer, int offset);
	
	/**
	 * Sets the offset of the viewed data object in buffer.
	 * 
	 * This method does <b>not</b> consider alignment! If 
	 * you want to set an aligned offset, then call:
	 * 
	 *<pre>
	 *view.offset(view.align(offset));
	 *</pre>
	 * 
	 * @param offset
	 * @see #offset()
	 * @see #align(int)
	 */
	void offset(int offset);
	
	/** 
	 * Calculates aligned offset for the given offset and returns the result. 
	 * <p>
	 * The alignment is performed in respect to the alignment currently set for this view.
	 * </p>
	 * @see #alignment()
	 */
	int align(int offset);
	
	/**
	 * Returns the offset into the buffer used by the view.
	 * @return currently used offset.
	 * @see #offset(int)
	 */
	int offset();
	
	
	/**
	 * @return returns the buffer currently used by this view.
	 */
	ByteBuffer buffer();

	/**
	 * Creates a clone of this view. The clone is <b>open</b> (see {@link #open(ByteBuffer)}) 
	 * and references the same buffer at the same offset and uses the same alignment,
	 * but it is independent of this (original) view. Thus, set operations on
	 * the clone do affect the same buffer but not the state of the 
	 * original view.
	 * <br/><br/>
	 * The clone is reusable (as every view is). You can open it with another buffer and different alignment (see {@link #open(ByteBuffer, int, int)}).
	 * <br/><br/>
	 * <b>Reminder: Call {@link #close()} when the clone is no longer needed!</b>
	 * 
	 * @return clone of this view.
	 */
	DataView clone();
	
	/**
	 * This method configures the given view to be a clone of this view (see {@link #clone()}.
	 * 
	 * @param view View to become a clone of this view.
	 * @return The given view, now configured to be a clone of this view.
	 */
	<V extends DataView> V clone(V view);
	
	/**
	 * As long as an opened view stays in memory, the attached buffer will stay
	 * memory as well. This method can be called to release the buffer.
	 * @see #open(ByteBuffer)
	 */
	void close();
	
	/**
	 * @return overall size of a single serialised data object of the type viewed by this view.
	 */
	int sizeof();
	
	/**
	 * This method calculates the distance ([bytes]) between two adjacent elements of the type 
	 * viewed by this view in respect to their alignment. It is the same as calling:
	 * <pre>
	 * this.align(this.sizeof())
	 * </pre>
	 * 
	 * @return distance between the start addresses of two adjacent elements of this type in respect to their alignment.
	 * @see #sizeof()
	 */
	int stride();
	
	/**
	 * @return returns the alignment set for this view.
	 */
	short alignment();
	
	/** Returns the Java type associated with this view. */
	Class<?> javatype();
	
}
