package org.cakelab.nio.view;

import java.io.Closeable;
import java.nio.ByteBuffer;
import java.util.Iterator;

/**
 * Iterator for a sequence of data objects of the same type in a buffer.
 * <p>
 * The iterator is instantiated with a data view. The given data view defines
 * the type of data objects to be iterated and the distance between two adjacent 
 * data objects ({@link DataView#stride()}). If the view is already opened, 
 * the iterator will also use the associated buffer.
 * The data view iterator is an alternative to array views, when the number 
 * of components is unknown.
 * </p>
 * 
 * <pre>
 * MyStruct data;
 * MyStructView myview = new MyStructView();
 * DataViewIterator&lt;MyStructView&gt; iterator = DataViewIterator.create(myview, buffer);
 * for (MyStructView v : iterator) {
 *    data = v.get();
 *    // .. do something with it
 * }
 * iterator.close(); // closes myview as well
 * 
 * </pre>
 * 
 * @author homac
 *
 * @param <V>
 */
public class DataViewIterator<V extends DataView> implements Iterable<V>, Iterator<V>, Closeable {

	private V view;
	private int stride;
	private int size;
	
	private int offset;
	private int end;

	/**
	 * Creates an iterator which starts in the buffer and at the offset, 
	 * the given view is pointing to, unless the view is not attached to a buffer.
	 * @param view view used to iterate over the stream
	 */
	public DataViewIterator(V view) {
		setView(view);
		if (view.buffer() != null) {
			setBoundaries(view.offset());
		}
	}
	
	/** creates an iterator which starts iterating over the buffer 
	 * attached to the view starting at the given offset. */
	public DataViewIterator(V view, int start) {
		this(view, view.buffer(), start);
	}

	/** Creates an iterator to iterate over the buffer using 
	 * the given view starting at buffer's current position. */
	public DataViewIterator(V view, ByteBuffer buffer) {
		this(view, buffer, buffer.position());
	}

	public DataViewIterator(V view, ByteBuffer buffer, int start) {
		setView(view);
		open(buffer, start);
	}
	
	
	private void setView(V view) {
		this.view = view;
		this.stride = view.stride();
		this.size = view.sizeof();
	}

	private void setBoundaries(int offset) {
		this.offset = offset;
		int end = view.buffer().capacity();
		int total = end-offset;
		int lastsize = (total)%stride;
		if (lastsize != 0 && lastsize < view.sizeof()) {
			end = ((total)/stride)*stride + view.sizeof() + offset;
		}
		this.end = end;
	}

	public void open(ByteBuffer buffer) {
		this.open(buffer, buffer.position());
	}

	public void open(ByteBuffer buffer, int offset) {
		this.view.open(buffer, offset);
		setBoundaries(offset);
	}

	public void reset() {
		offset = 0;
	}
	
	@Override
	public Iterator<V> iterator() {
		return this;
	}

	@Override
	public boolean hasNext() {
		return (offset + size) <= end;
	}

	@Override
	public V next() {
		assert (hasNext());
		view.offset(offset);
		offset += stride;
		return view;
	}

	/**
	 * Closes the given component view as well.
	 * 
	 * same as
	 * <pre>
	 * buf = view.buffer();
	 * buf.position(end-of-last)
	 * </pre>
	 * 
	 * {@inheritDoc}
	 * 
	 */
	public void close() {
		ByteBuffer buffer = view.buffer();
		if (buffer != null) {
			buffer.position(end);
			view.close();
		}
	}

	/**
	 * Creates a new data view iterator calling {@link #DataViewIterator(DataView, ByteBuffer)}
	 * @param view
	 * @param buf 
	 * @return new data view iterator instance.
	 */
	public static <T extends DataView> DataViewIterator<T> create(T view, ByteBuffer buf) {
		return new DataViewIterator<T>(view, buf);
	}

}
