package org.cakelab.nio.view;

/**
 * Interface of views on compound data objects (complex or array) which allow reuse 
 * (i.e. not immutable and not of primitive type).
 * <p>
 * Reusable data objects can be recycled in terms of object pooling, 
 * which makes use of the method {@link #get(Object)} beneficial. It allows 
 * general reuse of data objects in the context of read/write operations on 
 * buffers and thereby reduces GC overhead. 
 * </p>
 * <p>
 * For all reusable data objects, there will also be a method 
 * {@link #newDataInstance()}, which allows creation of an (uninitialised) 
 * object of that type. This improves efficiency of memory usage internally.
 * </p>
 * 
 * @author homac
 *
 * @param <T>
 */
public interface ReusableDataView<T> extends GenericDataView<T> {
	/**
	 * Get method writing the result in the given parameter 'result'.
	 * Reusing an existing object for the result allows to keep memory 
	 * allocation effort low.
	 * @param result Object, which receives the result.
	 * @return The given object. This return was added for convenience only.
	 * @see #get()
	 */
	T get(T result);
	
	/**
	 * Returns a new instance of the associated Java type of this view.
	 * Its the same as calling
	 * <pre>
	 * obj = new T();
	 * </pre>
	 * But with the advantage, that it will implicitly create new objects for all
	 * fields of a given class or elements of a given array, if those are reusable 
	 * as well (it's like a deep new).
	 * @return new instance of associated Java type.
	 */
	T newDataInstance();
}
