package org.cakelab.nio.view.annotations;

import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;

import org.cakelab.nio.view.DataView;


/**
 * This annotation declares specific attributes of 
 * structured types in respect to the native 
 * representation. 
 * <p>
 * The annotation is used on classes, which represent
 * the declaration of a structured type, or fields of 
 * those classes. It is required only, if the Java type, 
 * representing the struct, does not provide enough 
 * information, for example to specify the length of 
 * an array.
 * </p>
 * 
 * @author homac
 *
 */
@Retention(RetentionPolicy.RUNTIME)
public @interface nio {
	/**
	 * Default value for the attribute {@link #length()}.
	 * <p>
	 * It indicates, that the corresponding type or field
	 * refers to a single data item and not to a list of items.
	 * </p>
	 */
	final int LENGTH_NOT_SET = -1;
	
	/** 
	 * Default value for attribute {@link #view()}.
	 * <p>
	 * It indicates, that the attribute was not set 
	 * and the generator can use standard methods to
	 * access corresponding data.
	 * </p>
	 */
	final Class<? extends DataView> VIEW_NOT_SET = DataView.class;
	
	/** 
	 * Length of a list of data items. Default: {@link #LENGTH_NOT_SET} 
	 * <p>
	 * Number of components in an array or string for example. 
	 * </p>
	 * <p>
	 * This attribute accepts either 
	 * </p>
	 * <ul>
	 * <li><b>single integer values</b> (length=4) for arrays with one dimension or strings for example</li>
	 * <li><b>multiple integer values</b> (length={2,2}) for multi-dimensional arrays (e.g. matrices).</li>
	 * </ul>
	 * <p>
	 * The length attribute is typically used for array fields, such as:
	 * </p>
	 * <pre>
	 * &#64;nio(length=3)
	 * public float[] rgb;         // one dimensional
	 * &#64;nio(length={4,4})
	 * public float[][] mv_matrix; // two dimensional
	 * </pre>
	 *
	 * <p>
	 * A data item of type <code>java.lang.String</code> 
	 * is an example of a struct, which requires a length
	 * attribute to specify the length of the string. Thus, having an
	 * array of strings requires a length attribute, which specifies 
	 * the length of the array <em>and</em> the length of the strings
	 * contained in the array. This would result in a field 
	 * declaration such as this:
	 * </p>
	 * <pre>
	 * &#64;nio(length={4,16}, view=UTF8StringView.class, reusable=false)
	 * public String[] ids;        // one dimensional array of strings
	 * </pre>
	 * <p>This declares an array with 4 strings of length 16. 
	 * The second attribute specifies a particular view class 
	 * for strings with UTF-8 character encoding. A String also is
	 * non-reusable, which is indicated explicitly by the third 
	 * attribute here (even though it is not needed, see #{@link nio#reusable()}. 
	 * </p>
	 * <p>Analogously, a multi-dimensional array of strings would 
	 * look like this:
	 * </p>
	 * <pre>
	 * &#64;nio(length={2,2,16}, view=UTF8StringView.class, reusable=false)
	 * public String[][] devices;  // two dimensional array of strings
	 * </pre>
	 * <p>
	 * This declares an array of two arrays with two strings of length 16.
	 * The first n entries in the list of lengths provided to attribute <b>length</b>
	 * always applies to 
	 * the array (of length n) and the remaining entries will be forwarded as 
	 * parameters to the view class.
	 * </p>
	 * <p>
	 * A single string would have a declaration like this:
	 * </p>
	 * <pre>
	 * &#64;nio(length=256, view=UTF8StringView.class, reusable=false)
	 * public String filename;
	 * </pre>
	 */
	int[] length() default LENGTH_NOT_SET;
	
	/** 
	 * Data alignment. Default: <code>1</code> <br/>
	 * <h3>General Rules:</h3>
	 * <ul>
	 * <li>Given value has to be a power of two.</li>
	 * <li>Alignment can be specified for types, fields of a struct or elements of arrays.</li>
	 * <li>Alignment given for fields or elements of an array override declared alignment of their type.</li>
	 * <li>Alignment of a field will override alignment of its enclosing struct if it is larger. 
	 * 	   Thus, the struct will get the maximum value of the alignment of all of its fields.
	 * </li>
	 * <li>Accordingly, alignment of array elements will override the alignment of the array.</li>
	 * </ul>
	 * <h3>Explanation:</h3>
	 * <p>
	 * Alignment specifies at which addresses a particular field or data of 
	 * a particular type will be located in memory (i.e. buffer). The aligned
	 * address is calculated like this (pseudo code):
	 * </p>
	 * <pre>
	 * diff = (address % alignment);
	 * address = (diff == 0) ? address : address + (alignment-diff);
	 * </pre>
	 * <p>
	 * Thus, the address will always be shifted to the next address, which 
	 * satisfies <code>(address % alignment == 0)</code>.
	 * An alignment of 1 (default) allows corresponding data to be at any address. 
	 * </p>
	 * <h3>Declaration Examples</h3>
	 * <p>
	 * Alignment can be declared for structs (class annotation) or fields (field annotation).
	 * </p>
	 * <pre>
	 * &#64;nio(align=2)
	 * public class MyStruct {
	 *    &#64;nio(align=4)     // overrides struct alignment because (4 > 2)
	 *    public int value;
	 * }
	 * </pre>
	 * <p>
	 * If a struct has fields with larger alignment value than specified for the struct, 
	 * the alignment of the struct is simply ignored. 
	 * Overriding of alignment cannot conflict with the alignment which has 
	 * been overridden, because it is always larger and a power of two. 
	 * Thus, a larger alignment will always satisfy the smaller alignment, too.
	 * </p>
	 * <p>
	 * If a type of a field has a class annotation, with an alignment, it will use 
	 * that alignment unless a different alignment is specified for the field.
	 * </p>
	 * <pre>
	 * &#64;nio(align=2)
	 * public class MyOtherStruct {
	 *    &#64;nio(align=1)     // locally overrides alignment of type MyStruct
	 *    public MyStruct value;
	 * }
	 * </pre>
	 * <p>
	 * In case of an array, the alignment applies to the array elements only.
	 * </p>
	 * <pre>
	 * &#64;nio(length=8, align=4)     // short values with alignment 4
	 * public short[] value;
	 * </pre>
	 * <p>
	 * This example declares an array of short values, which are aligned to addresses
	 * which are a multiple of 4. Since short has a size of 2 in memory, the 
	 * buffer will contain 4 short values separated by 2 bytes each.
	 * </p>
	 * <p>
	 * If you want the array start to be aligned to 4, you need to insert a padding 
	 * value, like this:
	 * </p>
	 * <pre>
	 * &#64;nio(align=4)        // struct will have at least alignment of 4
	 * public PaddedStruct {
	 *    public char myvalue;  // counts 2 bytes (Character.BYTES)
	 *    public short padding; // add another 2 bytes to get 4 byte alignment for next field
	 *    &#64;nio(length=8)
	 *    public short[] value; // packed list of 8 short values (align=1)
	 * }
	 * </pre>
	 */
	int   align()  default 1;

	/**
	 * Specifies a custom view, to be used for a type, field 
	 * or the elements of an array. Default <b><code>VIEW_NOT_SET</code></b>.
	 * <p>
	 * This attribute can be used to specify a particular 
	 * view class to be used to access a given class or field of a class. 
	 * When applied on fields of any array type, the given view 
	 * is used on the elements of that array, not the array itself.
	 * </p>
	 */
	Class<? extends DataView> view() default DataView.class;
	
	/** 
	 * Indicates, that a data object of this type is reusable. Default <b><code>true</code></b>.
	 * <p>
	 * This attribute can be used on classes and fields of classes. It indicates, 
	 * whether an object of this type can be reused. This is generally not the case for 
	 * data objects of primitive type or immutable classes such as String, Integer etc..
	 * For all fields of reusable type the code generator will generate a getter method 
	 * which takes an object to store the data read from the byte buffer (see also {@link org.cakelab.nio.view.ReusableDataView}).
	 * </p>
	 * <p> <b>Note:</b> There are many standard types automatically detected to be non-reusable. Those are
	 * </p>
	 * <ul>
	 * <li>All primitive types such as <b><code>int</code></b>, <b><code>boolean</code></b></li>
	 * <li>all corresponding boxed types for primitive types, such as {@link java.lang.Integer} and {@link java.lang.Boolean} etc.</li>
	 * <li>and the class java.lang.String.</li>
	 * </ul>
	 * <p>Those types don't need to be explicitly declared to be non-reusable.
	 * </p>
	 * */
	boolean reusable() default true;

}
