package test;

/**
 * Simple value randomiser for test purposes.
 * <p> 
 * It works like a hash function (<code>h(v,s)</code>) which calculates for 
 * each given value <code>v</code> and seed <code>s</code> a different result (r). 
 * 
 * </p>
 * <pre>
 * r = h(v,s)
 * </pre>
 * <p>
 * The difference in results is meant to be small (r will be close to v) 
 * if s is small and large otherwise. This is necessary to design certain test cases.
 * </p>
 * <br/>
 * 
 * Given
 * <pre>
 * rng = new TestRandomizer();
 * </pre>
 * and
 * <pre>
 * x == y
 * </pre>
 * then 
 * <pre>
 * rng.get(x) == rng.get(y)
 * </pre> 
 * as well.
 * <br/>
 * But if
 * <pre>
 * a = rng.get(x);
 * rng.next();
 * </pre>
 * then
 * <pre> 
 * a != rng.get(y)
 * </pre>
 * in most cases.
 * 
 * @author homac
 *
 */
public class TestRandomizer {

	private int seed = 0;

	/** Sets seed value. <b>note:</b> considers only the least significant byte */
	public void setSeed(int seed) {
		this.seed = seed % Byte.MAX_VALUE;
	}

	/**
	 * returns current seed value
	 */
	public int getSeed() {
		return seed;
	}

	/**
	 * increments seed by one
	 */
	public void next() {
		setSeed(this.seed+1);
	}
	
	/** modifies given value based on seed */
	public byte get(byte b) {
		return (byte)((seed + b)&0xff);
	}

	/** modifies given value based on seed */
	public char get(char c) {
		int result = (seed + c) % Character.MAX_VALUE ;
		return (char)(result&0xffff);
	}

	/** modifies given value based on seed */
	public double get(double d) {
		return d + seed;
	}

	/** modifies given value based on seed */
	public float get(float f) {
		return f + seed;
	}

	/** modifies given value based on seed */
	public int get(int i) {
		return i + seed;
	}

	/** modifies given value based on seed */
	public long get(long l) {
		return l + seed;
	}

	/** modifies s based on seed */
	public short get(short s) {
		int result = (s + seed) % Short.MAX_VALUE;
		return (short) (result & 0xffff);
	}

	/** toggels given boolean based on seed */
	public boolean get(boolean b) {
		return seed%2 == 0 ? b : !b;
	}

	/** swaps randomly picked characters of the string. 
	 * seed influences selection of elements. */
	public String get(String string) {
		StringBuffer s = new StringBuffer(string);
		for (int i = 0; i < s.length(); i++) {
			int index = get(i) % s.length();
			s.setCharAt(i,string.charAt(index));
		}
		return s.toString();
	}

	/** swaps randomly picked elements of the array. 
	 * seed influences selection of elements. */
	public int[][] get(int[][] array) {
		for (int i = 0; i < array.length; i++) {
			int i2 = get(i) % array.length;
			for (int j = 0; j < array[i].length; j++) {
				int j2 = get(j) % array[i].length;
				int tmp = array[i][j];
				array[i][j] = array[i2][j2];
				array[i2][j2] = tmp;
			}
		}
		return array;
	}

	/** swaps randomly picked elements of the array. 
	 * seed influences selection of elements. */
	public short[] get(short[] array) {
		for (int i = 0; i < array.length; i++) {
			int i2 = get(i) % array.length;
			short tmp = array[i];
			array[i] = array[i2];
			array[i2] = tmp;
		}
		return array;
	}

	/** swaps randomly picked elements of the array. 
	 * seed influences selection of elements. */
	public long[][][] get(long[][][] array) {
		for (int i = 0; i < array.length; i++) {
			int i2 = get(i) % array.length;
			for (int j = 0; j < array[i].length; j++) {
				int j2 = get(j) % array[i].length;
				for (int k = 0; k < array[i][j].length; k++) {
					int k2 = get(k) % array[i][j].length;
					long tmp = array[i][j][k];
					array[i][j][k] = array[i2][j2][k2];
					array[i2][j2][k2] = tmp;
				}
			}
		}
		return array;
	}

	/** swaps randomly picked elements of the array. 
	 * seed influences selection of elements. */
	public char[] get(char[] array) {
		for (int i = 0; i < array.length; i++) {
			int i2 = get(i) % array.length;
			char tmp = array[i];
			array[i] = array[i2];
			array[i2] = tmp;
		}
		return array;
	}

	/** swaps randomly picked elements of the array. 
	 * seed influences selection of elements. */
	public float[] get(float[] array) {
		for (int i = 0; i < array.length; i++) {
			int i2 = get(i) % array.length;
			float tmp = array[i];
			array[i] = array[i2];
			array[i2] = tmp;
		}
		return array;
	}

	/** swaps randomly picked elements of the array. 
	 * seed influences selection of elements. */
	public double[] get(double[] array) {
		for (int i = 0; i < array.length; i++) {
			int i2 = get(i) % array.length;
			double tmp = array[i];
			array[i] = array[i2];
			array[i2] = tmp;
		}
		return array;
	}

}
