package test.array.view;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.nio.ByteBuffer;
import java.util.Arrays;

import org.cakelab.nio.view.array.CArrayView;
import org.cakelab.nio.view.array.intArrayView;
import org.cakelab.nio.view.array.impl.intArrayViewPacked;
import org.cakelab.nio.view.array.impl.intArrayViewUser;
import org.cakelab.nio.view.generated.test.array.complex.MyStructArrayView.ArrayElemView;
import org.cakelab.nio.view.string.StringView;

import test.TestRandomizer;
import test.array.complex.MyStructArray.ArrayElem;
import test.array.complex.TestMyStructArray;

public class TestArrayView {
	private static TestRandomizer rng = new TestRandomizer();

	
	public static void main(String[] args) throws InstantiationException, IllegalAccessException, ClassNotFoundException, NoSuchMethodException, SecurityException, IllegalArgumentException, InvocationTargetException, IOException {
		test();
	}
	
	public static void test() throws InstantiationException, IllegalAccessException, ClassNotFoundException, NoSuchMethodException, SecurityException, IllegalArgumentException, InvocationTargetException, IOException {
		test_intArrayViewPacked();
		test_intArrayViewSpaced(2);
		test_intArrayViewSpaced(8);
		test_intArrayViewUser(1);
		test_intArrayViewUser(2);
		test_intArrayViewUser(8);
		test_int24ArrayViewUser();
		testStringCArrayView();
		testCArrayView(1);
		testCArrayView(2);
		testCArrayView(32);
		testIntCArrayView();
		
		rng.next();
	}
	
	public static void test_intArrayViewPacked() {
		System.out.println("testing intArrayViewPacked");

		intArrayViewPacked ia = intArrayView.create(4);

		testIntArrayView(ia, Integer.BYTES, 4, 1);

	}
	public static void test_intArrayViewSpaced(int alignment) {
		System.out.println("testing intArrayViewSpaced (alignment=" + alignment + ")");
		
		int length = 4;
		intArrayView ia = intArrayView.create((short)alignment, length);
		testIntArrayView(ia, Integer.BYTES, length, alignment);
	}
	
	public static void test_intArrayViewUser(int alignment) {
		System.out.println("testing intArrayViewUser (alignment=" + alignment + ")");
		
		int length = 4;
		int24View elemView = new int24View((short)alignment);
		intArrayView ia = intArrayView.create(elemView, length);
		testIntArrayView(ia, elemView.sizeof(), length, alignment);
	}
	

	private static void testIntArrayView(intArrayView ia, int elemSize, int length, int alignment) {
		int stride = elemSize + (alignment-1) & ~(alignment-1);

		int[] data = new int[]{
			rng.get(0),
			rng.get(1),
			rng.get(2),
			rng.get(3)
		};

		assert length == data.length;
		
		assert data.length == ia.length();

		assert ia.alignment() == alignment;
		assert data.length == ia.length();
		assert ia.componentStride() == stride;
		assert ia.sizeof() == (stride * ia.length()) - (stride>elemSize?stride-elemSize:0);

		
		ia.open(ByteBuffer.allocate(ia.sizeof()));
		
		ia.set(data);
		int[] test = ia.get();
		assert Arrays.equals(data,test);
		for (int i = 0; i < ia.length(); i++) {
			assert ia.get(i) == data[i];
		}
		
		ia.set(4711, 2);
		assert ia.get(2) == 4711;
		
		ia.close();
	}

	public static void test_int24ArrayViewUser() {
		System.out.println("testing int24ArrayViewUser");

		intArrayView view = intArrayView.create(2);
		
		int24View vi24 = new int24View();
		
		intArrayViewUser<int24View> vi24a = intArrayView.create(vi24, 2);

		assert(view.length() == 2);
		assert(vi24a.length() == 2);
		
		int[] data;
		
		ByteBuffer buffer = ByteBuffer.allocate(view.sizeof());
		view.open(buffer);
		vi24a.open(buffer);
		
		int v = 1;
		
		view.set(v, 0);
		assert(view.get(0) == v);
		assert(vi24a.view(0).get() == 0);
		int24View eInt24View = vi24a.view(new int24View(), 0);
		assert(eInt24View.get() == 0);

		eInt24View.set(1);
		assert(vi24a.view(0).get() == 1);
		assert(eInt24View.get() == 1);
		
		
		
		
		v = Integer.MAX_VALUE;
		
		int i24_MAX_VALUE = int24View.toJavaInt(v);
		
		
		view.set(v, 0);
		assert(view.get(0) == v);
		assert(vi24a.view(0).get() == i24_MAX_VALUE);
		vi24a.view(eInt24View, 0);
		assert(eInt24View.get() == i24_MAX_VALUE);

		view.set(1, 1);

		int testI24 = vi24a.get(1);
		assert(view.get(1) != testI24);
		
		
		data = new int[]{rng.get(2),rng.get(3)};
		view.set(data);
		assert(Arrays.equals(view.get(),data)); 
		
		int[] test = new int[2];
		assert(Arrays.equals(view.get(test), data));


		view.close();
	}
	
	public static void testCArrayView(int alignment)  {
		System.out.println("testing CArrayView<Struct ...> (alignment=" + alignment + ")");
		
		int[] dimensions = new int[]{4,2};
		
		ArrayElem[][] data = new ArrayElem[][]{
			{new ArrayElem(1,2), new ArrayElem(3,4)},
			{new ArrayElem(5,6), new ArrayElem(7,8)},
			{new ArrayElem(9,10), new ArrayElem(11,12)},
			{new ArrayElem(13,14), new ArrayElem(15,16)},
		};
		ArrayElemView elemView = new ArrayElemView((short)alignment);
		CArrayView<ArrayElem, ArrayElemView> componentView = new CArrayView<ArrayElem, ArrayElemView>(elemView, dimensions[1]);
		CArrayView<ArrayElem[], CArrayView<ArrayElem, ArrayElemView>> arrayView_4_2 = new CArrayView<ArrayElem[], CArrayView<ArrayElem, ArrayElemView>>(componentView, dimensions[0]);
		
		assert(componentView.componentStride() == elemView.align(elemView.sizeof()));
		assert(arrayView_4_2.componentStride() == componentView.componentStride() * componentView.length());
		int aligndiff = componentView.componentStride() - elemView.sizeof();
		if (aligndiff < 0) aligndiff = 0; 
		assert(arrayView_4_2.sizeof() == dimensions[0]*dimensions[1]*componentView.componentStride()-(aligndiff));
		
		ByteBuffer buffer = ByteBuffer.allocate(arrayView_4_2.sizeof());
		
		arrayView_4_2.open(buffer);

		arrayView_4_2.set(data);

		ArrayElem[][] test = arrayView_4_2.get();
		assert(TestMyStructArray.equals(test, data));
		
		for (int i = 0; i < arrayView_4_2.length(); i++) {
			CArrayView<ArrayElem, ArrayElemView> array_2_view = arrayView_4_2.view(i);
			for (int j = 0; j < array_2_view.length(); j++) {
				elemView = (ArrayElemView) array_2_view.view(j);
				
				assert(TestMyStructArray.equals(data[i][j], elemView.get()));
				
				ArrayElem elem = array_2_view.get(j);
				assert(TestMyStructArray.equals(data[i][j], elem));
				assert(TestMyStructArray.equals(elem, elemView.get()));
				
			}
			ArrayElem[] array_2_component = arrayView_4_2.get(i);
			assert(TestMyStructArray.equals(data[i], array_2_component));
			assert(TestMyStructArray.equals(array_2_component, array_2_view.get()));
		}
		arrayView_4_2.close();
		
	}

	private static void testStringCArrayView() {
		System.out.println("testing CArrayView<StringView ...>");
		// primitive integer array with two dimensions [2][4]

		String[] data = new String[]{
			rng.get("hello"),
			rng.get("12345678"),
			rng.get("fantast"),
			rng.get("fox öäß")
		};

		StringView sv = new StringView(8);
		
		CArrayView<String, StringView> s1v = new CArrayView<String, StringView>(sv, 4);
		
		ByteBuffer buffer = ByteBuffer.allocate(s1v.sizeof());
		s1v.open(buffer);
		s1v.set(data);
		assert Arrays.equals(data, s1v.get());
		
		for (int i = 0; i < s1v.length(); i++) {
			assert s1v.get(i).equals(data[i]);
			assert s1v.view(i).get().equals(data[i]);
		}
		
		s1v.close();
	}

	private static void testIntCArrayView() {
		System.out.println("testing CArrayView<intArrayView ...>");
		// primitive integer array with two dimensions [2][4]
		
		intArrayViewPacked ia1v = intArrayView.create(4);
		
		CArrayView<int[],intArrayViewPacked> intCArrayView = new CArrayView<int[], intArrayViewPacked>(ia1v, 2);
		ByteBuffer buffer = ByteBuffer.allocate(intCArrayView.sizeof());
		intCArrayView.open(buffer);
		int[][] data = new int[][]{
			{
				rng.get(1),
				rng.get(2),
				rng.get(3),
				rng.get(4)
			},
			{
				rng.get(5),
				rng.get(6),
				rng.get(7),
				rng.get(8)
			}
		};
		intCArrayView.set(data);
		
		int[][] test = intCArrayView.get();
		assert(Arrays.deepEquals(data, test));
		for (int i = 0; i < intCArrayView.length(); i ++) {
			assert(Arrays.equals(data[i], intCArrayView.get(i)));
			intArrayViewPacked vi = intCArrayView.view(i);
			assert(Arrays.equals(data[i], vi.get()));
			assert(vi.length() == data[i].length);
			for (int j = 0; j < vi.length(); j++) {
				assert(data[i][j] == vi.get(j));
			}
		}
		intCArrayView.close();
	}


}
